# -*- coding: utf-8 -*-
#
#  method.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import commands.exceptions as exceptions
import types
import inspect
import sys
import utils

class Method:
    def __init__(self, method, name, parent):
        self.method = method
        self.real_name = name
        self.name = name.replace('_', '-')
        self.parent = parent
        self._func_props = None

    def __str__(self):
        return self.name

    def autocomplete_func(self):
        if hasattr(self.method, 'autocomplete'):
            return getattr(self.method, 'autocomplete')

        return None

    def accelerator(self):
        if hasattr(self.method, 'accelerator'):
            return getattr(self.method, 'accelerator')

        return None

    def args(self):
        fp = self.func_props()

        return fp.args, fp.varargs

    def func_props(self):
        if not self._func_props:
            # Introspect the function arguments
            self._func_props = utils.getargspec(self.method)

        return self._func_props

    def commands(self):
        return []

    def cancel(self, view):
        if self.parent:
            self.parent.cancel(view, self)

    def cancel_continuation(self, view):
        if self.parent:
            self.parent.continuation(view, self)

    def doc(self):
        if self.method.__doc__:
            return self.method.__doc__
        else:
            return ''

    def oneline_doc(self):
        return self.doc().split("\n")[0]

    def execute(self, argstr, words, entry, modifier, kk = {}):
        fp = self.func_props()

        kwargs = {'argstr': argstr, 'args': words, 'entry': entry, 'view': entry.view(), 'modifier': modifier, 'window': entry.view().get_toplevel()}
        oargs = list(fp.args)
        args = []
        idx = 0

        if fp.defaults:
            numdef = len(fp.defaults)
        else:
            numdef = 0

        for k in fp.args:
            if k in kwargs:
                args.append(kwargs[k])
                oargs.remove(k)
                del kwargs[k]
            elif idx >= len(words):
                if numdef < len(oargs):
                    raise exceptions.Execute('Invalid number of arguments (need %s)' % (oargs[0],))
            else:
                args.append(words[idx])
                oargs.remove(k)
                idx += 1

        # Append the rest if it can handle varargs
        if fp.varargs and idx < len(words):
            args.extend(words[idx:])

        if not fp.keywords:
            kwargs = {}

        for k in kk:
            kwargs[k] = kk[k]

        return self.method(*args, **kwargs)

    def __lt__(self, other):
        if isinstance(other, Method):
            return self.name < other.name
        else:
            return self.name < other

# vi:ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  entry.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, GLib, Gdk, Gtk
import cairo
import os
import re
import inspect
import sys

import commander.commands as commands
import commands.completion
import commands.module
import commands.method
import commands.exceptions
import commands.accel_group

import commander.utils as utils

from history import History
from info import Info
from xml.sax import saxutils
import traceback

class Entry(Gtk.EventBox):
    __gtype_name__ = "CommanderEntry"

    def __init__(self, view):
        Gtk.EventBox.__init__(self)
        self._view = view

        self.set_visible_window(False)

        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 3)
        hbox.show()
        hbox.set_border_width(3)

        # context for the view
        self._entry = Gtk.Entry()
        self._entry.set_has_frame(False)
        self._entry.set_name('gedit-commander-entry')
        self._entry.show()

        css = Gtk.CssProvider()
        css.load_from_data(bytes("""
@binding-set terminal-like-bindings {
    unbind "<Control>A";

    bind "<Control>W" { "delete-from-cursor" (word-ends, -1) };
    bind "<Control>A" { "move-cursor" (buffer-ends, -1, 0) };
    bind "<Control>U" { "delete-from-cursor" (display-line-ends, -1) };
    bind "<Control>K" { "delete-from-cursor" (display-line-ends, 1) };
    bind "<Control>E" { "move-cursor" (buffer-ends, 1, 0) };
    bind "Escape" { "delete-from-cursor" (display-lines, 1) };
}

GtkEntry#gedit-commander-entry {
    gtk-key-bindings: terminal-like-bindings;

    /* Override background to anything. This is weird, but doing this we can
       then in code use widget.override_background to set the color dynamically
       to the same color as the gedit view */
    background: transparent;
    border-width: 0;
    box-shadow: 0 0 transparent;
    transition: none;
}
""", 'utf-8'))

        # FIXME: remove hardcopy of 600 (GTK_STYLE_PROVIDER_PRIORITY_APPLICATION)
        # https://bugzilla.gnome.org/show_bug.cgi?id=646860
        self._entry.get_style_context().add_provider(css, 600)

        self._prompt_label = Gtk.Label(label='<b>&gt;&gt;&gt;</b>', use_markup=True)
        self._prompt_label.show()

        self._entry.connect('focus-out-event', self.on_entry_focus_out)
        self._entry.connect('key-press-event', self.on_entry_key_press)

        self._history = History(os.path.join(GLib.get_user_config_dir(), 'gedit/commander/history'))
        self._prompt = None

        self._accel_group = None

        hbox.pack_start(self._prompt_label, False, False, 0)
        hbox.pack_start(self._entry, True, True, 0)

        self.copy_style_from_view()
        self.view_style_updated_id = self._view.connect('style-updated', self.on_view_style_updated)

        self.add(hbox)
        self.attach()
        self._entry.grab_focus()

        self._wait_timeout = 0
        self._info_window = None

        self.connect('destroy', self.on_destroy)
        self.connect_after('size-allocate', self.on_size_allocate)
        self.view_draw_id = self._view.connect_after('draw', self.on_draw)

        self._history_prefix = None
        self._suspended = None
        self._handlers = [
            [0, Gdk.KEY_Up, self.on_history_move, -1],
            [0, Gdk.KEY_Down, self.on_history_move, 1],
            [None, Gdk.KEY_Return, self.on_execute, None],
            [None, Gdk.KEY_KP_Enter, self.on_execute, None],
            [0, Gdk.KEY_Tab, self.on_complete, None],
            [0, Gdk.KEY_ISO_Left_Tab, self.on_complete, None]
        ]

        self._re_complete = re.compile('("((?:\\\\"|[^"])*)"?|\'((?:\\\\\'|[^\'])*)\'?|[^\s]+)')
        self._command_state = commands.Commands.State()

    def on_view_style_updated(self, widget):
        self.copy_style_from_view()

    def get_border_color(self):
        color = self.get_background_color().copy()
        color.red = 1 - color.red
        color.green = 1 - color.green
        color.blue = 1 - color.blue
        color.alpha = 0.5

        return color

    def get_background_color(self):
        context = self._view.get_style_context()
        return context.get_background_color(Gtk.StateFlags.NORMAL)
    
    def get_foreground_color(self):
        context = self._view.get_style_context()
        return context.get_color(Gtk.StateFlags.NORMAL)

    def get_font(self):
        context = self._view.get_style_context()
        return context.get_font(Gtk.StateFlags.NORMAL)

    def copy_style_from_view(self, widget=None):
        if widget != None:
            context = self._view.get_style_context()
            font = context.get_font(Gtk.StateFlags.NORMAL)

            widget.override_color(Gtk.StateFlags.NORMAL, self.get_foreground_color())
            widget.override_font(self.get_font())
        else:
            if self._entry:
                self.copy_style_from_view(self._entry)

            if self._prompt_label:
                self.copy_style_from_view(self._prompt_label)

    def view(self):
        return self._view

    def on_size_allocate(self, widget, alloc):
        alloc = self.get_allocation()
        self._view.set_border_window_size(Gtk.TextWindowType.BOTTOM, alloc.height)

        win = self._view.get_window(Gtk.TextWindowType.BOTTOM)
        self.set_size_request(win.get_width(), -1)

        # NOTE: we need to do this explicitly somehow, otherwise the window
        # size will not be updated unless something else happens, not exactly
        # sure what. This might be caused by the multi notebook, or custom
        # animation layouting?
        self._view.get_parent().resize_children()

    def attach(self):
        # Attach ourselves in the text view, and position just above the
        # text window
        win = self._view.get_window(Gtk.TextWindowType.TEXT)
        alloc = self.get_allocation()

        self._view.set_border_window_size(Gtk.TextWindowType.BOTTOM, max(alloc.height, 1))
        self._view.add_child_in_window(self, Gtk.TextWindowType.BOTTOM, 0, 0)

        win = self._view.get_window(Gtk.TextWindowType.BOTTOM)

        self.show()
        self.set_size_request(win.get_width(), -1)

    def on_entry_focus_out(self, widget, evnt):
        if self._entry.get_sensitive():
            self.destroy()

    def on_entry_key_press(self, widget, evnt):
        state = evnt.state & Gtk.accelerator_get_default_mod_mask()
        text = self._entry.get_text()

        if evnt.keyval == Gdk.KEY_Escape:
            if self._info_window:
                if self._suspended:
                    self._suspended.resume()

                if self._info_window:
                    self._info_window.destroy()

                self._entry.set_sensitive(True)
            elif self._accel_group:
                self._accel_group = self._accel_group.parent

                if not self._accel_group or not self._accel_group.parent:
                    self._entry.set_editable(True)
                    self._accel_group = None

                self.prompt()
            elif text:
                self._entry.set_text('')
            elif self._command_state:
                self._command_state.clear()
                self.prompt()
            else:
                self._view.grab_focus()
                self.destroy()

            return True

        if state or self._accel_group:
            # Check if it should be handled by the accel group
            group = self._accel_group

            if not self._accel_group:
                group = commands.Commands().accelerator_group()

            accel = group.activate(evnt.keyval, state)

            if isinstance(accel, commands.accel_group.AccelGroup):
                self._accel_group = accel
                self._entry.set_text('')
                self._entry.set_editable(False)
                self.prompt()

                return True
            elif isinstance(accel, commands.accel_group.AccelCallback):
                self._entry.set_editable(True)
                self.run_command(lambda: accel.activate(self._command_state, self))
                return True

        if not self._entry.get_editable():
            return True

        for handler in self._handlers:
            if (handler[0] == None or handler[0] == state) and evnt.keyval == handler[1] and handler[2](handler[3], state):
                return True

        if self._info_window and self._info_window.empty():
            self._info_window.destroy()

        self._history_prefix = None
        return False

    def on_history_move(self, direction, modifier):
        pos = self._entry.get_position()

        self._history.update(self._entry.get_text())

        if self._history_prefix == None:
            if len(self._entry.get_text()) == pos:
                self._history_prefix = self._entry.get_chars(0, pos)
            else:
                self._history_prefix = ''

        if self._history_prefix == None:
            hist = ''
        else:
            hist = self._history_prefix

        next = self._history.move(direction, hist)

        if next != None:
            self._entry.set_text(next)
            self._entry.set_position(-1)

        return True

    def prompt(self, pr=''):
        self._prompt = pr

        if self._accel_group != None:
            pr = '<i>%s</i>' % (saxutils.escape(self._accel_group.full_name()),)

        if not pr:
            pr = ''
        else:
            pr = ' ' + pr

        self._prompt_label.set_markup('<b>&gt;&gt;&gt;</b>%s' % pr)

    def make_info(self):
        if self._info_window == None:
            self._info_window = Info(self)
            self._info_window.show()

            self._info_window.connect('destroy', self.on_info_window_destroy)

    def on_info_window_destroy(self, widget):
        self._info_window = None

    def info_show(self, text='', use_markup=False):
        self.make_info()
        self._info_window.add_lines(text, use_markup)

    def info_status(self, text):
        self.make_info()
        self._info_window.status(text)

    def info_add_action(self, stock, callback, data=None):
        self.make_info()
        return self._info_window.add_action(stock, callback, data)

    def command_history_done(self):
        self._history.add(self._entry.get_text())
        self._history_prefix = None
        self._entry.set_text('')

    def on_wait_cancel(self):
        if self._suspended:
            self._suspended.resume()

        if self._cancel_button:
            self._cancel_button.destroy()

        if self._info_window and self._info_window.empty():
            self._info_window.destroy()
            self._entry.grab_focus()
            self._entry.set_sensitive(True)

    def _show_wait_cancel(self):
        self._cancel_button = self.info_add_action(Gtk.STOCK_STOP, self.on_wait_cancel)
        self.info_status('<i>Waiting to finish...</i>')

        self._wait_timeout = 0
        return False

    def _complete_word_match(self, match):
        for i in (3, 2, 0):
            if match.group(i) != None:
                return [match.group(i), match.start(i), match.end(i)]

    def on_suspend_resume(self):
        if self._wait_timeout:
            GLib.source_remove(self._wait_timeout)
            self._wait_timeout = 0
        else:
            self._cancel_button.destroy()
            self._cancel_button = None
            self.info_status(None)

        self._entry.set_sensitive(True)
        self.command_history_done()

        if self._entry.props.has_focus or (self._info_window and not self._info_window.empty()):
            self._entry.grab_focus()

        self.on_execute(None, 0)

    def ellipsize(self, s, size):
        if len(s) <= size:
            return s

        mid = (size - 4) / 2
        return s[:mid] + '...' + s[-mid:]

    def destroy(self):
        self.hide()
        Gtk.EventBox.destroy(self)

    def run_command(self, cb):
        self._suspended = None

        try:
            ret = cb()
        except Exception as e:
            self.command_history_done()
            self._command_state.clear()

            self.prompt()

            # Show error in info
            self.info_show('<b><span color="#f66">Error:</span></b> ' + saxutils.escape(str(e)), True)

            if not isinstance(e, commands.exceptions.Execute):
                self.info_show(self.format_trace(), False)

            return None

        mod = sys.modules['commander.commands.result']

        if ret == mod.Result.SUSPEND:
            # Wait for it...
            self._suspended = ret
            ret.register(self.on_suspend_resume)

            self._wait_timeout = GLib.timeout_add(500, self._show_wait_cancel)
            self._entry.set_sensitive(False)
        else:
            self.command_history_done()
            self.prompt('')

            if ret == mod.Result.PROMPT:
                self.prompt(ret.prompt)
            elif (ret == None or ret == mod.HIDE) and not self._prompt and (not self._info_window or self._info_window.empty()):
                self._command_state.clear()
                self._view.grab_focus()
                self.destroy()
            else:
                self._entry.grab_focus()

        return ret

    def format_trace(self):
        tp, val, tb = sys.exc_info()

        origtb = tb

        thisdir = os.path.dirname(__file__)

        # Skip frames up until after the last entry.py...
        while True:
            filename = tb.tb_frame.f_code.co_filename

            dname = os.path.dirname(filename)

            if not dname.startswith(thisdir):
                break

            tb = tb.tb_next

        msg = traceback.format_exception(tp, val, tb)
        r = ''.join(msg[0:-1])

        # This is done to prevent cyclic references, see python
        # documentation on sys.exc_info
        del origtb

        return r

    def on_execute(self, dummy, modifier):
        if self._info_window and not self._suspended:
            self._info_window.destroy()

        text = self._entry.get_text().strip()
        words = list(self._re_complete.finditer(text))
        wordsstr = []

        for word in words:
            spec = self._complete_word_match(word)
            wordsstr.append(spec[0])

        if not wordsstr and not self._command_state:
            self._entry.set_text('')
            return

        self.run_command(lambda: commands.Commands().execute(self._command_state, text, words, wordsstr, self, modifier))

        return True

    def on_complete(self, dummy, modifier):
        # First split all the text in words
        text = self._entry.get_text()
        pos = self._entry.get_position()

        words = list(self._re_complete.finditer(text))
        wordsstr = []

        for word in words:
            spec = self._complete_word_match(word)
            wordsstr.append(spec[0])

        # Find out at which word the cursor actually is
        # Examples:
        #  * hello world|
        #  * hello| world
        #  * |hello world
        #  * hello wor|ld
        #  * hello  |  world
        #  * "hello world|"
        posidx = None

        for idx in range(0, len(words)):
            spec = self._complete_word_match(words[idx])

            if words[idx].start(0) > pos:
                # Empty space, new completion
                wordsstr.insert(idx, '')
                words.insert(idx, None)
                posidx = idx
                break
            elif spec[2] == pos:
                # At end of word, resume completion
                posidx = idx
                break
            elif spec[1] <= pos and spec[2] > pos:
                # In middle of word, do not complete
                return True

        if posidx == None:
            wordsstr.append('')
            words.append(None)
            posidx = len(wordsstr) - 1

        # First word completes a command, if not in any special 'mode'
        # otherwise, relay completion to the command, or complete by advice
        # from the 'mode' (prompt)
        cmds = commands.Commands()

        if not self._command_state and posidx == 0:
            # Complete the first command
            ret = commands.completion.command(words=wordsstr, idx=posidx)
        else:
            complete = None
            realidx = posidx

            if not self._command_state:
                # Get the command first
                cmd = commands.completion.single_command(wordsstr, 0)
                realidx -= 1

                ww = wordsstr[1:]
            else:
                cmd = self._command_state.top()
                ww = wordsstr

            if cmd:
                complete = cmd.autocomplete_func()

            if not complete:
                return True

            # 'complete' contains a dict with arg -> func to do the completion
            # of the named argument the command (or stack item) expects
            args, varargs = cmd.args()

            # Remove system arguments
            s = ['argstr', 'args', 'entry', 'view']
            args = list(filter(lambda x: not x in s, args))

            if realidx < len(args):
                arg = args[realidx]
            elif varargs:
                arg = '*'
            else:
                return True

            if not arg in complete:
                return True

            func = complete[arg]

            try:
                spec = utils.getargspec(func)

                if not ww:
                    ww = ['']

                kwargs = {
                    'words': ww,
                    'idx': realidx,
                    'view': self._view
                }

                if not spec.keywords:
                    for k in list(kwargs.keys()):
                        if not k in spec.args:
                            del kwargs[k]

                ret = func(**kwargs)
            except Exception as e:
                # Can be number of arguments, or return values or simply buggy
                # modules
                print(e)
                traceback.print_exc()
                return True

        if not ret or not ret[0]:
            return True

        res = ret[0]
        completed = ret[1]

        if len(ret) > 2:
            after = ret[2]
        else:
            after = ' '

        # Replace the word
        if words[posidx] == None:
            # At end of everything, just append
            spec = None

            self._entry.insert_text(completed, self._entry.get_text_length())
            self._entry.set_position(-1)
        else:
            spec = self._complete_word_match(words[posidx])

            self._entry.delete_text(spec[1], spec[2])
            self._entry.insert_text(completed, spec[1])
            self._entry.set_position(spec[1] + len(completed))

        if len(res) == 1:
            # Full completion
            lastpos = self._entry.get_position()

            if not isinstance(res[0], commands.module.Module) or not res[0].commands():
                if words[posidx] and after == ' ' and (words[posidx].group(2) != None or words[posidx].group(3) != None):
                    lastpos = lastpos + 1

                self._entry.insert_text(after, lastpos)
                self._entry.set_position(lastpos + 1)
            elif completed == wordsstr[posidx] or not res[0].method:
                self._entry.insert_text('.', lastpos)
                self._entry.set_position(lastpos + 1)

            if self._info_window:
                self._info_window.destroy()
        else:
            # Show popup with completed items
            if self._info_window:
                self._info_window.clear()

            ret = []

            for x in res:
                if isinstance(x, commands.method.Method):
                    ret.append('<b>' + saxutils.escape(x.name) + '</b> (<i>' + x.oneline_doc() + '</i>)')
                else:
                    ret.append(str(x))

            self.info_show("\n".join(ret), True)

        return True

    def on_draw(self, widget, ct):
        win = widget.get_window(Gtk.TextWindowType.BOTTOM)

        if not Gtk.cairo_should_draw_window(ct, win):
            return False

        Gtk.cairo_transform_to_window(ct, widget, win)

        color = self.get_border_color()
        width = win.get_width()

        ct.set_source_rgba(color.red, color.green, color.blue, color.alpha)
        ct.move_to(0, 0)
        ct.line_to(width, 0)
        ct.stroke()

        return False

    def on_destroy(self, widget):
        self._view.set_border_window_size(Gtk.TextWindowType.BOTTOM, 0)
        self._view.disconnect(self.view_style_updated_id)
        self._view.disconnect(self.view_draw_id)

        if self._info_window:
            self._info_window.destroy()

        self._history.save()

# vi:ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  __init__.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import os
import sys

path = os.path.dirname(__file__)

if not path in sys.path:
    sys.path.insert(0, path)

from windowactivatable import WindowActivatable
import commander.commands as commands
from gi.repository import GObject, GLib, Gedit

class CommanderPlugin(GObject.Object, Gedit.AppActivatable):
    __gtype_name__ = "CommanderPlugin"

    app = GObject.property(type=Gedit.App)

    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        self._path = os.path.dirname(__file__)

        if not self._path in sys.path:
            sys.path.insert(0, self._path)

        commands.Commands().set_dirs([
            os.path.join(GLib.get_user_config_dir(), 'gedit/commander/modules'),
            os.path.join(self.plugin_info.get_data_dir(), 'modules')
        ])

    def deactivate(self):
        commands.Commands().stop()

        if self._path in sys.path:
            sys.path.remove(self._path)

# vi:ex:ts=4:et



#!/usr/bin/env python
# -- coding: utf-8 --
#
# Copyright © 2011 Collabora Ltd.
#            By Seif Lotfy <seif@lotfy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

from gi.repository import GObject, Gedit
from .dashboard import Dashboard

class DashboardWindowActivatable(GObject.Object, Gedit.WindowActivatable):

    window = GObject.property(type=Gedit.Window)

    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        self.status_bar = self.window.get_statusbar()
        self._handlers = [self.window.connect_after("tab-added", self._add_tab),
            self.window.connect_after("active-tab-changed", self._toggle_status_bar),
            self.window.connect_after("destroy", lambda x: self.status_bar.show())]

    def _toggle_status_bar(self, window, tab):
        self.status_bar.show()
        for w in tab.get_children():
            if type(w) == Dashboard:
                self.status_bar.hide()
                break

    def _add_tab(self, window, tab):

        notebook = tab.get_parent()

        doc = tab.get_document()
        uri = doc.get_uri_for_display()

        if not uri.startswith("/"):
            notebook.get_tab_label(tab).get_children()\
                    [0].get_children()[0].get_children()[1].hide()
            self.status_bar.hide()
            d = None
            if str(type(notebook.get_tab_label(tab))) == \
                "<class 'gi.types.__main__.GeditTabLabel'>":
                label = notebook.get_tab_label(tab).get_children()\
                    [0].get_children()[0].get_children()[2]
                text = label.get_text()
                label.set_text("Getting Started")
                def show_doc():
                    notebook.get_tab_label(tab).get_children()\
                        [0].get_children()[0].get_children()[1].show()
                    tab.remove(d)
                    tab.get_children()[0].show()
                    tab.grab_focus()
                    if label.get_text() == "Getting Started":
                        label.set_text(text)
                    self.status_bar.show()
                tab.get_children()[0].hide()
                d = Dashboard(self.window, show_doc)
                tab.pack_start(d, True, True, 0)
                doc.connect("loaded", lambda x, y: show_doc())
                d.search.set_receives_default(True)
                self.window.set_default(d.search)
                d.search.grab_focus()

    def do_deactivate(self):
        tab = self.window.get_active_tab()
        if tab:
            notebook = tab.get_parent()
            for tab in notebook.get_children():
                for w in tab.get_children():
                    if type(w) == Dashboard:
                        tab.remove(w)
                    else:
                        w.show()
                doc = tab.get_document()
                label = notebook.get_tab_label(tab).get_children()\
                    [0].get_children()[0].get_children()[2]
                label.set_text(doc.get_short_name_for_display())

        for handler in self._handlers:
           self.window.disconnect(handler)
        self.window = None

# ex:ts=4:et:



# -*- coding: utf-8 -*-
#
#  set.py - set commander module
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import commander.commands as commands
import commander.commands.exceptions

import types
from gi.repository import GtkSource

__commander_module__ = True

def _complete_options(words, idx):
    ret = []

    gb = globals()

    for k in gb:
        if type(gb[k]) == types.FunctionType and not k.startswith('_'):
            ret.append(k.replace('_', '-'))

    ret.sort()
    return commands.completion.words(ret)(words, idx)

def _complete_language(words, idx):
    manager = GtkSource.LanguageManager.get_default()
    ids = manager.get_language_ids()
    ids.append('none')
    ids.sort()

    return commands.completion.words(ids)(words, idx)

def _complete_use_spaces(words, idx):
    return commands.completion.words(['yes', 'no'])(words, idx)

def _complete_draw_spaces(words, idx):
    ret = ['none', 'all', 'tabs', 'newlines', 'nbsp', 'spaces']
    return commands.completion.words(ret)(words, idx)

def _complete_value(words, idx):
    # Depends a bit on the option
    ret, completion = _complete_options(words, idx - 1)

    if not ret:
        return None

    completer = '_complete_' + ret[0].replace('-', '_')
    gb = globals()

    if completer in gb:
        return gb[completer](words[1:], idx - 1)
    else:
        return None

@commands.autocomplete(option=_complete_options, value=_complete_value)
def __default__(view, option, value):
    """Set gedit option: set &lt;option&gt; &lt;value&gt;

Sets a gedit option, such as document language, or indenting"""

    option = option.replace('-', '_')
    gb = globals()

    if option in gb and type(gb[option]) == types.FunctionType:
        return gb[option](view, value)
    else:
        raise commander.commands.exceptions.Execute('Invalid setting: ' + option)

@commands.autocomplete(language=_complete_language)
def language(view, language=None):
    """Set document language: set.language &lt;language&gt;

Set the document language to the language with the specified id"""
    if not language or language == 'none':
        view.get_buffer().set_language(None)
        return False

    manager = GtkSource.LanguageManager.get_default()
    lang = manager.get_language(language)

    if lang:
        view.get_buffer().set_language(lang)
        return False
    else:
        raise commander.commands.exceptions.Execute('Invalid language: ' + language)

def tab_width(view, width):
    """Set document tab width: set.tab-width &lt;width&gt;

Set the document tab width"""

    try:
        width = int(width)
    except:
        raise commander.commands.exceptions.Execute("Invalid tab width: " + str(width))

    if width <= 0:
        raise commander.commands.exceptions.Execute("Invalid tab width: " + str(width))

    view.set_tab_width(width)
    return False

tab_size = tab_width

@commands.autocomplete(value=_complete_use_spaces)
def use_spaces(view, value):
    """Use spaces instead of tabs: set.use-spaces &lt;yes/no&gt;

Set to true/yes to use spaces instead of tabs"""

    setting = value in ('yes', 'true', '1')
    view.set_insert_spaces_instead_of_tabs(setting)

    return False

@commands.autocomplete({'*': _complete_draw_spaces})
def draw_spaces(view, *args):
    """Draw spaces: set.draw-spaces &lt;none/all/tabs/newlines/nbsp/spaces&gt;

Set what kind of spaces should be drawn. Multiple options can be defined, e.g.
for drawing spaces and tabs: <i>set.draw-spaces space tab</i>"""
    m = {
        'none': 0,
        'all': GtkSource.DrawSpacesFlags.ALL,
        'tabs': GtkSource.DrawSpacesFlags.TAB,
        'newlines': GtkSource.DrawSpacesFlags.NEWLINE,
        'nbsp': GtkSource.DrawSpacesFlags.NBSP,
        'spaces': GtkSource.DrawSpacesFlags.SPACE
    }

    flags = 0

    for arg in args:
        for a in m:
            if a.startswith(arg):
                flags = flags | m[a]

    view.set_draw_spaces(flags)
    return False

# vi:ex:ts=4:et



# -*- coding: utf-8 -*-

#  Copyright (C) 2013 - Garrett Regier
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330,
#  Boston, MA 02111-1307, USA.

from gi.repository import GLib, GObject, Gio, Gtk, Gedit, Ggit

import weakref


class GitWindowActivatable(GObject.Object, Gedit.WindowActivatable):
    window = GObject.property(type=Gedit.Window)

    windows = weakref.WeakValueDictionary()

    def __init__(self):
        super().__init__()

        Ggit.init()

        self.view_activatables = weakref.WeakSet()

        self.files = {}
        self.monitors = {}

        self.repo = None
        self.tree = None

    @classmethod
    def register_view_activatable(cls, view_activatable):
        window = view_activatable.view.get_toplevel()

        if window not in cls.windows:
            return None

        window_activatable = cls.windows[window]

        window_activatable.view_activatables.add(view_activatable)
        view_activatable.connect('notify::status',
                                 window_activatable.notify_status)

        return window_activatable

    def do_activate(self):
        # self.window is not set until now
        self.windows[self.window] = self

        self.bus = self.window.get_message_bus()

        self.files = {}
        self.file_names = {}
        self.monitors = {}

        self.window_signals = [
            self.window.connect('tab-removed', self.tab_removed),
            self.window.connect('focus-in-event', self.focus_in_event)
        ]

        self.bus_signals = [
            self.bus.connect('/plugins/filebrowser', 'root_changed',
                             self.root_changed, None),
            self.bus.connect('/plugins/filebrowser', 'inserted',
                             self.inserted, None),
            self.bus.connect('/plugins/filebrowser', 'deleted',
                             self.deleted, None)
        ]

        self.refresh();

    def do_deactivate(self):
        self.clear_monitors()

        for sid in self.window_signals:
            self.window.disconnect(sid)

        for sid in self.bus_signals:
            self.bus.disconnect(sid)

        self.files = {}
        self.file_names = {}
        self.repo = None
        self.tree = None
        self.window_signals = []
        self.bus_signals = []

        self.refresh();

    def refresh(self):
        if self.bus.is_registered('/plugins/filebrowser', 'refresh'):
            self.bus.send('/plugins/filebrowser', 'refresh')

    def notify_status(self, view_activatable, psepc):
        location = view_activatable.view.get_buffer().get_location()

        if location is not None:
            self.update_location(location)

    def tab_removed(self, window, tab):
        view = tab.get_view()
        location = view.get_buffer().get_location()

        if location is None:
            return

        # Need to remove the view activatable otherwise update_location()
        # will might use the view's status and not the file's actual status
        for view_activatable in self.view_activatables:
            if view_activatable.view == view:
                self.view_activatables.remove(view_activatable)
                break

        self.update_location(location)

    def focus_in_event(self, window, event):
        for view_activatable in self.view_activatables:
            view_activatable.update()

        for uri in self.files:
            self.update_location(Gio.File.new_for_uri(uri))

    def root_changed(self, bus, msg, data=None):
        self.clear_monitors()

        try:
            repo_file = Ggit.Repository.discover(msg.location)
            self.repo = Ggit.Repository.open(repo_file)
            head = self.repo.get_head()
            commit = self.repo.lookup(head.get_target(), Ggit.Commit.__gtype__)
            self.tree = commit.get_tree()

        except Exception as e:
            self.repo = None
            self.tree = None

        else:
            self.monitor_directory(msg.location)

    def inserted(self, bus, msg, data=None):
        self.files[msg.location.get_uri()] = msg.id
        self.file_names[msg.location.get_uri()] = msg.name

        self.update_location(msg.location)

        if msg.is_directory:
            self.monitor_directory(msg.location)

    def deleted(self, bus, msg, data=None):
        # File browser's deleted signal is broken
        return

        uri = msg.location.get_uri()

        del self.files[uri]
        del self.file_names[uri]

        if uri in self.monitors:
            self.monitors[uri].cancel()
            del self.monitors[uri]

    def update_location(self, location):
        if self.repo is None:
            return

        if location.get_uri() not in self.files:
            return

        status = None

        # First get the status from the open documents
        for view_activatable in self.view_activatables:
            doc_location = view_activatable.view.get_buffer().get_location()

            if doc_location is not None and doc_location.equal(location):
                status = view_activatable.status
                break

        try:
            git_status = self.repo.file_status(location)

        except Exception:
            pass

        else:
            # Don't use the view activatable's
            # status if the file is ignored
            if status is None or git_status & Ggit.StatusFlags.IGNORED:
                status = git_status

        markup = GLib.markup_escape_text(self.file_names[location.get_uri()]) 

        if status is not None:
            if status & Ggit.StatusFlags.INDEX_NEW or \
                    status & Ggit.StatusFlags.WORKING_TREE_NEW or \
                    status & Ggit.StatusFlags.INDEX_MODIFIED or \
                    status & Ggit.StatusFlags.WORKING_TREE_MODIFIED:
                markup = '<span weight="bold">%s</span>' % (markup)

            elif status & Ggit.StatusFlags.INDEX_DELETED or \
                    status & Ggit.StatusFlags.WORKING_TREE_DELETED:
                markup = '<span strikethrough="true">%s</span>' % (markup)

        self.bus.send('/plugins/filebrowser', 'set_markup',
                      id=self.files[location.get_uri()], markup=markup)

    def clear_monitors(self):
        for uri in self.monitors:
            self.monitors[uri].cancel()

        self.monitors = {}

    def monitor_directory(self, location):
        monitor = location.monitor(Gio.FileMonitorFlags.NONE, None)
        self.monitors[location.get_uri()] = monitor

        monitor.connect('changed', self.monitor_changed)

    def monitor_changed(self, monitor, file, other_file, event_type):
        # Only monitor for changes as the file browser will monitor
        # file created and deleted files and emit inserted and deleted
        if event_type == Gio.FileMonitorEvent.CHANGED:
            for f in (file, other_file):
                if f in self.files:
                    self.update_location(f)

# ex:ts=4:et:



# -*- coding: utf-8 -*-
#
#  entry.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import os

class History:
    def __init__(self, filename):
        self._filename = filename
        self._history = ['']
        self._ptr = 0

        self.load()

    def find(self, direction, prefix):
        ptr = self._ptr + direction

        while ptr >= 0 and ptr < len(self._history):
            if self._history[ptr].startswith(prefix):
                return ptr

            ptr += direction

        return -1

    def move(self, direction, prefix):
        next = self.find(direction, prefix)

        if next != -1:
            self._ptr = next
            return self._history[self._ptr]
        else:
            return None

    def up(self, prefix=''):
        return self.move(-1, prefix)

    def down(self, prefix=''):
        return self.move(1, prefix)

    def add(self, line):
        if line.strip() != '':
            if self._history[-1] != '':
                self._history.append(line)
            else:
                self._history[-1] = line

        if self._history[-1] != '':
            self._history.append('')

        self._ptr = len(self._history) - 1

    def update(self, line):
        self._history[self._ptr] = line

    def load(self):
        try:
            #FIXME: we should rewrite this
            self._history = list(map(lambda x: x.strip("\n"), open(self._filename, 'r').readlines()))
            self._history.append('')
            self._ptr = len(self._history) - 1
        except IOError:
            pass

    def save(self):
        try:
            os.makedirs(os.path.dirname(self._filename))
        except OSError:
            pass

        try:
            f = open(self._filename, 'w')

            if self._history[-1] == '':
                hist = self._history[:-1]
            else:
                hist = self._history

            f.writelines(map(lambda x: x + "\n", hist))
            f.close()
        except IOError:
            pass

# vi:ex:ts=4:et



# -*- coding: utf-8 -*-
#  Join lines plugin
#  This file is part of gedit
#
#  Copyright (C) 2006-2007 Steve Frécinaux, André Homeyer
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, Gtk, Gedit
import gettext
from gpdefs import *

try:
    gettext.bindtextdomain(GETTEXT_PACKAGE, GP_LOCALEDIR)
    _ = lambda s: gettext.dgettext(GETTEXT_PACKAGE, s)
except:
    _ = lambda s: s

ui_str = """
<ui>
  <menubar name="MenuBar">
    <menu name="EditMenu" action="Edit">
      <placeholder name="EditOps_5">
        <menuitem name="JoinLines" action="JoinLines"/>
        <menuitem name="SplitLines" action="SplitLines"/>
      </placeholder>
    </menu>
  </menubar>
</ui>
"""

class JoinLinesPlugin(GObject.Object, Gedit.WindowActivatable):
    __gtype_name__ = "JoinLinesPlugin"

    window = GObject.property(type=Gedit.Window)

    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        self._insert_menu()

    def do_deactivate(self):
        self._remove_menu()

    def do_update_state(self):
        view = self.window.get_active_view()
        self._action_group.set_sensitive(view is not None and \
                                         view.get_editable())

    def _remove_menu(self):
        manager = self.window.get_ui_manager()
        manager.remove_ui(self._ui_id)
        manager.remove_action_group(self._action_group)
        manager.ensure_update()

    def _insert_menu(self):
        manager = self.window.get_ui_manager()
        self._action_group = Gtk.ActionGroup(name="GeditJoinLinesPluginActions")
        self._action_group.add_actions(
            [("JoinLines", None, _("_Join Lines"), "<Ctrl>J",
              _("Join the selected lines"),
              lambda a, w: join_lines(w)),
             ("SplitLines", None, _('_Split Lines'), "<Shift><Ctrl>J",
              _("Split the selected lines"),
              lambda a, w: split_lines(w))],
            self.window)

        manager.insert_action_group(self._action_group)
        self._ui_id = manager.add_ui_from_string(ui_str)

def join_lines(window):
    doc = window.get_active_document()
    if doc is None:
        return

    doc.begin_user_action()

    # If there is a selection use it, otherwise join the
    # next line
    try:
        start, end = doc.get_selection_bounds()
    except ValueError:
        start = doc.get_iter_at_mark(doc.get_insert())
        end = start.copy()
        end.forward_line()

    end_mark = doc.create_mark(None, end)

    if not start.ends_line():
        start.forward_to_line_end()

    # Include trailing spaces in the chunk to be removed
    while start.backward_char() and start.get_char() in ('\t', ' '):
        pass
    start.forward_char()

    while doc.get_iter_at_mark(end_mark).compare(start) == 1:
        end = start.copy()
        while end.get_char() in ('\r', '\n', ' ', '\t'):
            end.forward_char()
        doc.delete(start, end)

        doc.insert(start, ' ')
        start.forward_to_line_end()

    doc.delete_mark(end_mark)
    doc.end_user_action()

def split_lines(window):
    view = window.get_active_view()
    if view is None:
        return

    doc = view.get_buffer()

    width = view.get_right_margin_position()
    tabwidth = view.get_tab_width()

    doc.begin_user_action()

    try:
        # get selection bounds
        start, end = doc.get_selection_bounds()

        # measure indent until selection start
        indent_iter = start.copy()
        indent_iter.set_line_offset(0)
        indent = ''
        while indent_iter.get_offset() != start.get_offset():
            if indent_iter.get_char() == '\t':
                indent = indent + '\t'
            else:
                indent = indent + ' '
            indent_iter.forward_char()
    except ValueError:
        # select from start to line end
        start = doc.get_iter_at_mark(doc.get_insert())
        start.set_line_offset(0)
        end = start.copy()
        if not end.ends_line():
            end.forward_to_line_end()

        # measure indent of line
        indent_iter = start.copy()
        indent = ''
        while indent_iter.get_char() in (' ', '\t'):
            indent = indent + indent_iter.get_char()
            indent_iter.forward_char()

    end_mark = doc.create_mark(None, end)

    # ignore first word
    previous_word_end = start.copy()
    forward_to_word_start(previous_word_end)
    forward_to_word_end(previous_word_end)

    while 1:
        current_word_start = previous_word_end.copy()
        forward_to_word_start(current_word_start)

        current_word_end = current_word_start.copy()
        forward_to_word_end(current_word_end)

        if ord(current_word_end.get_char()) and \
           doc.get_iter_at_mark(end_mark).compare(current_word_end) >= 0:

            word_length = current_word_end.get_offset() - \
                          current_word_start.get_offset()

            doc.delete(previous_word_end, current_word_start)

            line_offset = get_line_offset(current_word_start, tabwidth) + word_length
            if line_offset > width - 1:
                doc.insert(current_word_start, '\n' + indent)
            else:
                doc.insert(current_word_start, ' ')

            previous_word_end = current_word_start.copy()
            previous_word_end.forward_chars(word_length)
        else:
            break

    doc.delete_mark(end_mark)
    doc.end_user_action()

def get_line_offset(text_iter, tabwidth):
    offset_iter = text_iter.copy()
    offset_iter.set_line_offset(0)

    line_offset = 0
    while offset_iter.get_offset() < text_iter.get_offset():
        char = offset_iter.get_char()
        if char == '\t':
            line_offset += tabwidth
        else:
            line_offset += 1
        offset_iter.forward_char()

    return line_offset

def forward_to_word_start(text_iter):
    char = text_iter.get_char()
    while ord(char) and (char in (' ', '\t', '\n', '\r')):
        text_iter.forward_char()
        char = text_iter.get_char()

def forward_to_word_end(text_iter):
    char = text_iter.get_char()
    while ord(char) and (not (char in (' ', '\t', '\n', '\r'))):
        text_iter.forward_char()
        char = text_iter.get_char()

# ex:ts=4:et:



# -*- coding: utf-8 -*-
#
#  completion.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import commander.commands as commands
import bisect
import sys
import os
import re

from xml.sax import saxutils

__all__ = ['command', 'filename']

def _common_prefix_part(first, second):
    length = min(len(first), len(second))

    for i in range(0, length):
        if first[i] != second[i]:
            return first[:i]

    return first[:length]

def common_prefix(args, sep=None):
    # A common prefix can be something like
    # first: some-thing
    # second: sho-tar
    # res: s-t
    args = list(args)

    if not args:
        return ''

    if len(args) == 1:
        return str(args[0])

    first = str(args[0])
    second = str(args[1])

    if not sep:
        ret = _common_prefix_part(first, second)
    else:
        first = first.split(sep)
        second = second.split(sep)
        ret = []

        for i in range(0, min(len(first), len(second))):
            ret.append(_common_prefix_part(first[i], second[i]))

        ret = sep.join(ret)

    del args[0]
    args[0] = ret

    return common_prefix(args, sep)

def _expand_commands(cmds):
    if not cmds:
        cmds.extend(commands.Commands().modules())
        return

    old = list(cmds)
    del cmds[:]

    # Expand 'commands' to all the respective subcommands

    for cmd in old:
        for c in cmd.commands():
            bisect.insort(cmds, c)

def _filter_command(cmd, subs):
    parts = cmd.name.split('-')

    if len(subs) > len(parts):
        return False

    for i in range(0, len(subs)):
        if not parts[i].startswith(subs[i]):
            return False

    return True

def _filter_commands(cmds, subs):
    # See what parts of cmds still match the parts in subs
    idx = bisect.bisect_left(cmds, subs[0])
    ret = []

    while idx < len(cmds):
        if not cmds[idx].name.startswith(subs[0]):
            break

        if _filter_command(cmds[idx], subs):
            ret.append(cmds[idx])

        idx += 1

    return ret

def single_command(words, idx):
    ret = command(words, idx)

    if not ret:
        return None

    ret[0] = list(filter(lambda x: x.method, ret[0]))

    if not ret[0]:
        return None

    return ret[0][0]

def command(words, idx):
    s = words[idx].strip()

    parts = s.split('.')
    cmds = []

    for i in parts:
        # Expand all the parents to their child commands
        _expand_commands(cmds)

        if not cmds:
            return None

        subs = i.split('-')
        cmds = _filter_commands(cmds, subs)

        if not cmds:
            return None

    if not cmds:
        return None

    if len(parts) == 1:
        completed = common_prefix(cmds)
    else:
        completed = '.'.join(parts[0:-1]) + '.' + common_prefix(cmds, '-')

    return [cmds, completed]

def _file_color(path):
    if os.path.isdir(path):
        format = '<span color="#799ec6">%s</span>'
    else:
        format = '%s'

    return format % (saxutils.escape(os.path.basename(path)),)

def _sort_nicely(l):
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]

    l.sort(key=alphanum_key)

def filename(words, idx, view):
    prefix = os.path.dirname(words[idx])
    partial = os.path.expanduser(words[idx])

    doc = view.get_buffer()

    if not doc.is_untitled():
        root = os.path.dirname(doc.get_location().get_path())
    else:
        root = os.path.expanduser('~/')

    if not os.path.isabs(partial):
        partial = os.path.join(root, partial)

    dirname = os.path.dirname(partial)

    try:
        files = os.listdir(dirname)
    except OSError:
        return None

    base = os.path.basename(partial)
    ret = []
    real = []

    for f in files:
        if f.startswith(base) and (base or not f.startswith('.')):
            real.append(os.path.join(dirname, f))
            ret.append(os.path.join(prefix, f))

    _sort_nicely(real)

    if len(ret) == 1:
        if os.path.isdir(real[0]):
            after = '/'
        else:
            after = ' '

        return ret, ret[0], after
    else:
        return list(map(lambda x: _file_color(x), real), common_prefix(ret))

def words(ret):
    def decorator(words, idx):
        rr = list(filter(lambda x: x.startswith(words[idx]), ret))
        return rr, common_prefix(rr)

    return decorator

# vi:ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  goto.py - goto commander module
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import os

import commander.commands as commands
import commander.commands.completion
import commander.commands.result
import commander.commands.exceptions

__commander_module__ = True

def __default__(view, line, column=1):
    """Goto line number"""

    buf = view.get_buffer()
    ins = buf.get_insert()
    citer = buf.get_iter_at_mark(ins)

    try:
        if line.startswith('+'):
            linnum = citer.get_line() + int(line[1:])
        elif line.startswith('-'):
            linnum = citer.get_line() - int(line[1:])
        else:
            linnum = int(line) - 1

        column = int(column) - 1
    except ValueError:
        raise commander.commands.exceptions.Execute('Please specify a valid line number')

    linnum = min(max(0, linnum), buf.get_line_count() - 1)
    citer = buf.get_iter_at_line(linnum)

    column = min(max(0, column), citer.get_chars_in_line() - 1)

    citer = buf.get_iter_at_line(linnum)
    citer.forward_chars(column)

    buf.place_cursor(citer)
    view.scroll_to_iter(citer, 0.0, True, 0, 0.5)

    return commander.commands.result.HIDE

# vi:ex:ts=4:et



#!/usr/bin/python
# -*- coding: utf-8 -*-

# This file is part of the Gedit Synctex plugin.
#
# Copyright (C) 2010 Jose Aliste <jose.aliste@gmail.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

import dbus, subprocess, time

RUNNING, CLOSED = range(2)

EV_DAEMON_PATH = "/org/gnome/evince/Daemon"
EV_DAEMON_NAME = "org.gnome.evince.Daemon"
EV_DAEMON_IFACE = "org.gnome.evince.Daemon"

EVINCE_PATH = "/org/gnome/evince/Evince"
EVINCE_IFACE = "org.gnome.evince.Application"

EV_WINDOW_IFACE = "org.gnome.evince.Window"



class EvinceWindowProxy:
    """A DBUS proxy for an Evince Window."""
    daemon = None
    bus = None

    def __init__(self, uri, spawn = False, logger = None):
        self._log = logger
        self.uri = uri
        self.spawn = spawn
        self.status = CLOSED
        self.source_handler = None
        self.dbus_name = ''
        self._handler = None
        try:
            if EvinceWindowProxy.bus is None:
                EvinceWindowProxy.bus = dbus.SessionBus()

            if EvinceWindowProxy.daemon is None:
                EvinceWindowProxy.daemon = EvinceWindowProxy.bus.get_object(EV_DAEMON_NAME,
                                                EV_DAEMON_PATH,
                                                follow_name_owner_changes=True)
            EvinceWindowProxy.bus.add_signal_receiver(self._on_doc_loaded, signal_name="DocumentLoaded", 
                                                      dbus_interface = EV_WINDOW_IFACE, 
                                                      sender_keyword='sender')
            self._get_dbus_name(False)

        except dbus.DBusException:
            if self._log:
                self._log.debug("Could not connect to the Evince Daemon")

    def _on_doc_loaded(self, uri, **keyargs):
        if uri == self.uri and self._handler is None:
            self.handle_find_document_reply(keyargs['sender'])
        
    def _get_dbus_name(self, spawn):
        EvinceWindowProxy.daemon.FindDocument(self.uri,spawn,
                     reply_handler=self.handle_find_document_reply,
                     error_handler=self.handle_find_document_error,
                     dbus_interface = EV_DAEMON_IFACE)

    def handle_find_document_error(self, error):
        if self._log:
            self._log.debug("FindDocument DBus call has failed")

    def handle_find_document_reply(self, evince_name):
        if self._handler is not None:
            handler = self._handler
        else:
            handler = self.handle_get_window_list_reply
        if evince_name != '':
            self.dbus_name = evince_name
            self.status = RUNNING
            self.evince = EvinceWindowProxy.bus.get_object(self.dbus_name, EVINCE_PATH)
            self.evince.GetWindowList(dbus_interface = EVINCE_IFACE,
                          reply_handler = handler,
                          error_handler = self.handle_get_window_list_error)

    def handle_get_window_list_error (self, e):
        if self._log:
            self._log.debug("GetWindowList DBus call has failed")

    def handle_get_window_list_reply (self, window_list):
        if len(window_list) > 0:
            window_obj = EvinceWindowProxy.bus.get_object(self.dbus_name, window_list[0])
            self.window = dbus.Interface(window_obj,EV_WINDOW_IFACE)
            self.window.connect_to_signal("Closed", self.on_window_close)
            self.window.connect_to_signal("SyncSource", self.on_sync_source)
        else:
            #That should never happen. 
            if self._log:
                self._log.debug("GetWindowList returned empty list")


    def set_source_handler (self, source_handler):
        self.source_handler = source_handler

    def on_window_close(self):
        self.window = None
        self.status = CLOSED

    def on_sync_source(self, input_file, source_link, timestamp):
        if self.source_handler is not None:
            self.source_handler(input_file, source_link, timestamp)

    def SyncView(self, input_file, data, time):
        if self.status == CLOSED:
            if self.spawn:
                self._tmp_syncview = [input_file, data, time];
                self._handler = self._syncview_handler
                self._get_dbus_name(True)
        else:
            self.window.SyncView(input_file, data, time,  dbus_interface = "org.gnome.evince.Window")

    def _syncview_handler(self, window_list):
        self.handle_get_window_list_reply(window_list)

        if self.status == CLOSED: 
            return False
        self.window.SyncView(self._tmp_syncview[0],self._tmp_syncview[1], self._tmp_syncview[2], dbus_interface="org.gnome.evince.Window")
        del self._tmp_syncview
        self._handler = None
        return True

## This file can be used as a script to support forward search and backward search in vim.
## It should be easy to adapt to other editors. 
##  evince_dbus  pdf_file  line_source input_file
if __name__ == '__main__':
    import dbus.mainloop.glib, sys, os, logging
    from gi.repository import GObject

    def print_usage():
        print('''
The usage is evince_dbus output_file line_number input_file from the directory of output_file.
''')
        sys.exit(1)

    if len(sys.argv)!=4:
        print_usage()
    try:
        line_number = int(sys.argv[2])
    except ValueError:
        print_usage()

    output_file = sys.argv[1]
    input_file  = sys.argv[3]
    path_output  = os.getcwd() + '/' + output_file
    path_input   = os.getcwd() + '/' + input_file

    if not os.path.isfile(path_output):
        print_usage()

    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    logger = logging.getLogger("evince_dbus")
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    ch.setFormatter(formatter)

    logger.addHandler(ch)    
    a = EvinceWindowProxy('file://' + path_output, True,logger=logger)
    
    def sync_view(ev_window, path_input, line_number):
        ev_window.SyncView (path_input, (line_number, 1),0)

    GObject.timeout_add(400, sync_view, a, path_input, line_number)
    loop = GObject.MainLoop()
    loop.run() 
# ex:ts=4:et:



#!/usr/bin/env python
# -- coding: utf-8 --
#
# Copyright © 2011 Collabora Ltd.
#            By Seif Lotfy <seif@lotfy.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, Gtk, Gio, GLib, GdkPixbuf, GtkSource
import os
import urllib.parse

thumb_cache = {}
pixbuf_cache = {}

ICON_SIZES = SIZE_NORMAL, SIZE_LARGE = ((128, 128), (256, 256))

# Thumbview and Timelineview sizes
SIZE_THUMBVIEW = [(48, 36), (96, 72), (150, 108), (256, 192)]
SIZE_TIMELINEVIEW = [(16, 12), (32, 24), (64, 48), (128, 96)]
SIZE_TEXT_TIMELINEVIEW = ["small", "medium", "large", "x-large"]
SIZE_TEXT_THUMBVIEW = ["x-small", "small", "large", "xx-large"]

GRID_ITEM_COUNT = 7

no_pixbuf = Gtk.IconTheme.get_default().load_icon("gtk-file", 48, 0)


def uri_exists(uri):
    return uri.startswith("note://") or \
        os.path.exists(urllib.parse.unquote(str(uri[7:])))


def make_icon_frame(pixbuf, blend=False, border=1, color=0x000000ff):
    """creates a new Pixmap which contains 'pixbuf' with a border at given
    size around it."""
    if not pixbuf:
        return pixbuf
    result = GdkPixbuf.Pixbuf.new(pixbuf.get_colorspace(), True,
        pixbuf.get_bits_per_sample(), pixbuf.get_width(),
        pixbuf.get_height())
    result.fill(color)
    pixbuf.copy_area(border, border, pixbuf.get_width() - (border * 2),
        pixbuf.get_height() - (border * 2), result, border, border)
    return result


def create_text_thumb(gfile, size=None, threshold=2):
    """ tries to use source_view to get a thumbnail of a text file """
    content = gfile.get_content().split("\n")
    new_content = []
    last_line = ""
    for line in content:
        x = line.strip()
        if not (x.startswith("/*") or x.startswith("*")\
            or x.startswith("#") or x.startswith("import ")\
            or x.startswith("from ") or x.startswith("#include ")\
            or x.startswith("#define") or x.startswith("using")\
            or x == "") and not last_line.endswith("\\"):
            new_content.append(line)
        else:
            last_line = line
    missing_lines = 30 - len(new_content)
    if missing_lines > 0:
        for i in range(missing_lines):
            new_content.append(" \n")
    content = "\n".join(new_content)
    lang_manager = GtkSource.LanguageManager.get_default()
    lang = lang_manager.guess_language(
        gfile.subject.get_property("uri")[7:], None)
    if lang:
        buf = GtkSource.Buffer.new_with_language(lang)
    else:
        buf = GtkSource.Buffer()
    buf.set_highlight_syntax(True)
    buf.set_text(content)
    source_view = GtkSource.View.new_with_buffer(buf)
    overlay = Gtk.Overlay()
    overlay.add(Gtk.EventBox())
    overlay.add_overlay(source_view)
    event_box = Gtk.EventBox()
    event_box.add(overlay)
    return event_box


def __crop_pixbuf(pixbuf, x, y, size=SIZE_LARGE):
    """ returns a part of the given pixbuf as new one """
    result = GdkPixbuf.Pixbuf.new(pixbuf.get_colorspace(),
        True,
        pixbuf.get_bits_per_sample(),
        size[0], size[1])
    pixbuf.copy_area(x, y, x + size[0], y + size[1], result, 0, 0)
    return result


def get_pixbuf_from_cache(gicon, callback, size=48):
    name = gicon.to_string()
    if name in pixbuf_cache:
        callback(pixbuf_cache[name])
    else:
        theme = Gtk.IconTheme.get_default()
        # blocks: do something with it
        gtk_icon = theme.lookup_by_gicon(gicon, size,
            Gtk.IconLookupFlags.FORCE_SIZE)
        if gtk_icon:
            icon = gtk_icon.load_icon()
        else:
            icon = None
        pixbuf_cache[name] = icon
        callback(icon)


def get_icon(item, callback, size=48):
    uri = item.get_property("uri")
    f = Gio.file_new_for_uri(uri)

    def query_info_callback(gfile, result, cancellable):
        info = f.query_info_finish(result)

        def pixbuf_loaded_cb(pixbuf):
            if pixbuf is None:
                return
            icon = pixbuf.scale_simple(size * pixbuf.get_width()\
                / pixbuf.get_height(), size, 2)
            callback(icon)


        gicon = info.get_icon()
        get_pixbuf_from_cache(gicon, pixbuf_loaded_cb, size)


    if f.is_native():
        attr = Gio.FILE_ATTRIBUTE_STANDARD_ICON + ","\
            + Gio.FILE_ATTRIBUTE_THUMBNAIL_PATH

        def cancel_cb():
            pass

        f.query_info_async(attr, Gio.FileQueryInfoFlags.NONE,
            GLib.PRIORITY_DEFAULT, None, query_info_callback, None)




# -*- coding: utf-8 -*-

#  Copyright (C) 2013 - Ignacio Casal Quinteiro
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330,
#  Boston, MA 02111-1307, USA.

from gi.repository import GLib, GObject, Gtk, Gedit, Ggit
from .diffrenderer import DiffType, DiffRenderer
from .windowactivatable import GitWindowActivatable

import difflib


class LineContext:
    __slots__ = ('removed_lines', 'line_type')

    def __init__(self):
        self.removed_lines = []
        self.line_type = DiffType.NONE


class GitViewActivatable(GObject.Object, Gedit.ViewActivatable):
    view = GObject.property(type=Gedit.View)

    status = GObject.property(type=Ggit.StatusFlags,
                              default=Ggit.StatusFlags.CURRENT)

    def __init__(self):
        super().__init__()

        self.diff_timeout = 0
        self.file_contents_list = None
        self.file_context = None

    def do_activate(self):
        GitWindowActivatable.register_view_activatable(self)

        self.diff_renderer = DiffRenderer()
        self.gutter = self.view.get_gutter(Gtk.TextWindowType.LEFT)

        self.view_signals = [
            self.view.connect('notify::buffer', self.on_notify_buffer),
            self.view.connect('focus-in-event', self.update_location)
        ]

        self.buffer = None
        self.on_notify_buffer(self.view)

    def do_deactivate(self):
        if self.diff_timeout != 0:
            GLib.source_remove(self.diff_timeout)

        self.disconnect_buffer()
        self.buffer = None

        self.disconnect_view()
        self.gutter.remove(self.diff_renderer)

    def disconnect(self, obj, signals):
        for sid in signals:
            obj.disconnect(sid)

        signals[:] = []

    def disconnect_buffer(self):
        self.disconnect(self.buffer, self.buffer_signals)

    def disconnect_view(self):
        self.disconnect(self.view, self.view_signals)

    def on_notify_buffer(self, view, gspec=None):
        if self.diff_timeout != 0:
            GLib.source_remove(self.diff_timeout)

        if self.buffer:
            self.disconnect_buffer()

        self.buffer = view.get_buffer()

        # The changed signal is connected to in update_location()
        self.buffer_signals = [
            self.buffer.connect('loaded', self.update_location),
            self.buffer.connect('saved', self.update_location)
        ]

        # We wait and let the loaded signal call
        # update_location() as the buffer is currently empty

    def update_location(self, *args):
        self.location = self.buffer.get_location()
        if self.location is None:
            return

        try:
            repo_file = Ggit.Repository.discover(self.location)
            repo = Ggit.Repository.open(repo_file)
            head = repo.get_head()
            commit = repo.lookup(head.get_target(), Ggit.Commit.__gtype__)
            tree = commit.get_tree()

        except Exception:
            # Not a git repository
            if self.file_contents_list is not None:
                self.file_contents_list = None
                self.gutter.remove(self.diff_renderer)
                self.diff_renderer.set_file_context({})
                self.buffer.disconnect(self.buffer_signals.pop())

            return

        if self.file_contents_list is None:
            self.gutter.insert(self.diff_renderer, 40)
            self.buffer_signals.append(self.buffer.connect('changed',
                                                           self.update))

        try:
            relative_path = repo.get_workdir().get_relative_path(self.location)

            entry = tree.get_by_path(relative_path)
            file_blob = repo.lookup(entry.get_id(), Ggit.Blob.__gtype__)
            file_contents = file_blob.get_raw_content()
            self.file_contents_list = file_contents.decode('utf-8').splitlines()

            # Remove the last empty line added by gedit automatically
            last_item = self.file_contents_list[-1]
            if last_item[-1:] == '\n':
                self.file_contents_list[-1] = last_item[:-1]

        except Exception:
            # New file in a git repository
            self.file_contents_list = []

        self.update()

    def update(self, *unused):
        # We don't let the delay accumulate
        if self.diff_timeout != 0:
            return

        # Do the initial diff without a delay
        if self.file_context is None:
            self.on_diff_timeout()

        else:
            n_lines = self.buffer.get_line_count()
            delay = min(10000, 200 * (n_lines // 2000 + 1))

            self.diff_timeout = GLib.timeout_add(delay,
                                                 self.on_diff_timeout)

    def on_diff_timeout(self):
        self.diff_timeout = 0

        # Must be a new file
        if not self.file_contents_list:
            self.status = Ggit.StatusFlags.WORKING_TREE_NEW

            n_lines = self.buffer.get_line_count()
            if len(self.diff_renderer.file_context) == n_lines:
                return False

            line_context = LineContext()
            line_context.line_type = DiffType.ADDED
            file_context = dict(zip(range(1, n_lines + 1),
                                    [line_context] * n_lines))

            self.diff_renderer.set_file_context(file_context)
            return False

        start_iter, end_iter = self.buffer.get_bounds()
        src_contents = start_iter.get_visible_text(end_iter)
        src_contents_list = src_contents.splitlines()

        # GtkTextBuffer does not consider a trailing "\n" to be text
        if len(src_contents_list) != self.buffer.get_line_count():
            src_contents_list.append('')

        diff = difflib.unified_diff(self.file_contents_list,
                                    src_contents_list, n=0)

        # Skip the first 2 lines: ---, +++
        try:
            next(diff)
            next(diff)

        except StopIteration:
            # Nothing has changed
            self.status = Ggit.StatusFlags.CURRENT

        else:
            self.status = Ggit.StatusFlags.WORKING_TREE_MODIFIED

        file_context = {}
        for line_data in diff:
            if line_data[0] == '@':
                for token in line_data.split():
                    if token[0] == '+':
                        hunk_point = int(token.split(',', 1)[0])
                        line_context = LineContext()
                        break

            elif line_data[0] == '-':
                if line_context.line_type == DiffType.NONE:
                    line_context.line_type = DiffType.REMOVED

                line_context.removed_lines.append(line_data[1:])

                # No hunk point increase
                file_context[hunk_point] = line_context

            elif line_data[0] == '+':
                if line_context.line_type == DiffType.NONE:
                    line_context.line_type = DiffType.ADDED
                    file_context[hunk_point] = line_context

                elif line_context.line_type == DiffType.REMOVED:
                    # Why is this the only one that does
                    # not add it to file_context?

                    line_context.line_type = DiffType.MODIFIED

                else:
                    file_context[hunk_point] = line_context

                hunk_point += 1

        # Occurs when all of the original content is deleted
        if 0 in file_context:
            for i in reversed(list(file_context.keys())):
                file_context[i + 1] = file_context[i]
                del file_context[i]

        self.file_context = file_context
        self.diff_renderer.set_file_context(file_context)
        return False

# ex:ts=4:et:



# -*- coding: utf-8 -*-
#
#  transparentwindow.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, Gdk, Gtk, Gedit
import cairo

class TransparentWindow(Gtk.Window):
    __gtype_name__ = "CommanderTransparentWindow"

    def __init__(self, lvl=Gtk.WindowType.TOPLEVEL):
        Gtk.Window.__init__(self,
                            type=lvl,
                            decorated=False,
                            app_paintable=True,
                            skip_pager_hint=True,
                            skip_taskbar_hint=True)

        self.set_events(Gdk.EventMask.ALL_EVENTS_MASK)
        self.set_rgba()

    def set_rgba(self):
        visual = self.get_screen().get_rgba_visual()

        if not visual:
            visual = self.get_screen().get_system_visual()

        self.set_visual(visual)

    def do_screen_changed(self, prev):
        super(TransparentWindow, self).do_screen_changed(prev)

        self.set_rgba()

    def background_color(self):
        return Gdk.RGBA(0, 0, 0, 0.8)

    def background_shape(self, ct, alloc):
        ct.rectangle(0, 0, alloc.width, alloc.height)

    def draw_background(self, ct, widget=None):
        if widget == None:
            widget = self

        ct.set_operator(cairo.OPERATOR_SOURCE)
        alloc = widget.get_allocation()

        ct.rectangle(0, 0, alloc.width, alloc.height)
        ct.set_source_rgba(0, 0, 0, 0)
        ct.fill()

        color = self.background_color()
        self.background_shape(ct, alloc)

        ct.set_source_rgba(color.red, color.green, color.blue, color.alpha)
        ct.fill()

    def do_draw(self, ct):
        self.draw_background(ct)
        Gtk.Window.do_draw(self, ct)

        return False

# vi:ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  windowhelper.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, Gtk, Gedit
from entry import Entry
from info import Info
from gpdefs import *

try:
    gettext.bindtextdomain(GETTEXT_PACKAGE, GP_LOCALEDIR)
    _ = lambda s: gettext.dgettext(GETTEXT_PACKAGE, s);
except:
    _ = lambda s: s

ui_str = """
<ui>
  <menubar name="MenuBar">
    <menu name="EditMenu" action="Edit">
      <placeholder name="EditOps_5">
        <menuitem name="CommanderEditMode" action="CommanderModeAction"/>
      </placeholder>
    </menu>
  </menubar>
</ui>
"""

class WindowActivatable(GObject.Object, Gedit.WindowActivatable):
    __gtype_name__ = "CommanderWindowActivatable"

    window = GObject.property(type=Gedit.Window)

    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        self._entry = None
        self._view = None

        self.install_ui()

    def do_deactivate(self):
        self.uninstall_ui()

    def do_update_state(self):
        pass

    def install_ui(self):
        manager = self.window.get_ui_manager()

        self._action_group = Gtk.ActionGroup("GeditCommanderPluginActions")
        self._action_group.add_toggle_actions([('CommanderModeAction', None,
                                               _('Commander Mode'), '<Ctrl>period',
                                               _('Start commander mode'), self.on_commander_mode)])

        manager.insert_action_group(self._action_group, -1)
        self._merge_id = manager.add_ui_from_string(ui_str)

    def uninstall_ui(self):
        manager = self.window.get_ui_manager()
        manager.remove_ui(self._merge_id)
        manager.remove_action_group(self._action_group)

        manager.ensure_update()

    def on_commander_mode(self, action, user_data=None):
        view = self.window.get_active_view()

        if not view:
            return False

        if action.get_active():
            if not self._entry or view != self._view:
                self._entry = Entry(view)
                self._entry.connect('destroy', self.on_entry_destroy)

            self._entry.grab_focus()
            self._view = view
        elif self._entry:
            self._entry.destroy()
            self._view = None

        return True

    def on_entry_destroy(self, widget, user_data=None):
        self._entry = None
        self._action_group.get_action('CommanderModeAction').set_active(False)

# vi:ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  __init__.py - Text size plugin
#
#  Copyright (C) 2008 - Konstantin Mikhaylov <jtraub.devel@gmail.com>
#  Copyright (C) 2009 - Wouter Bolsterlee <wbolster@gnome.org>
#  Copyright (C) 2010 - Ignacio Casal Quinteiro <icq@gnome.org>
#  Copyright (C) 2010 - Jesse van den Kieboom <jessevdk@gnome.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, Gtk, Gdk, Gedit
from .documenthelper import DocumentHelper
import gettext
from gpdefs import *

try:
    gettext.bindtextdomain(GETTEXT_PACKAGE, GP_LOCALEDIR)
    _ = lambda s: gettext.dgettext(GETTEXT_PACKAGE, s);
except:
    _ = lambda s: s

# UI manager snippet to add menu items to the View menu
ui_str = """
<ui>
  <menubar name="MenuBar">
    <menu name="ViewMenu" action="View">
      <placeholder name="ViewOps_2">
        <separator/>
        <menuitem name="LargerText" action="LargerTextAction"/>
        <menuitem name="SmallerText" action="SmallerTextAction"/>
        <menuitem name="NormalSize" action="NormalSizeAction"/>
        <separator/>
      </placeholder>
    </menu>
  </menubar>
</ui>
"""

class TextSizePlugin(GObject.Object, Gedit.WindowActivatable):
    __gtype_name__ = "TextSizePlugin"

    window = GObject.property(type=Gedit.Window)

    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        self._views  = {}

        # Insert menu items
        self._insert_menu()

        # Insert document helpers
        for view in self.window.get_views():
            self.add_document_helper(view)

        self.window.connect('tab-added', self.on_tab_added)
        self.window.connect('tab-removed', self.on_tab_removed)

        self._accel_group = Gtk.AccelGroup()
        self.window.add_accel_group(self._accel_group)

        self._proxy_callback_map = {
            'LargerTextAction': self.on_larger_text_accel,
            'SmallerTextAction': self.on_smaller_text_accel,
            'NormalSizeAction': self.on_normal_size_accel
        }

        self._proxy_mapping = {}
        self._init_proxy_accels()
        self._accel_map_handler_id = Gtk.AccelMap.get().connect('changed', self.on_accel_map_changed)

    def _install_proxy(self, action):
        if not isinstance(action, Gtk.Action):
            action = self._action_group.get_action(str(action))

        if not action:
            return

        entry = Gtk.AccelMap.lookup_entry(action.get_accel_path())

        if not entry:
            return

        mapping = {
            Gdk.KEY_plus: Gdk.KEY_KP_Add,
            Gdk.KEY_KP_Add: Gdk.KEY_plus,
            Gdk.KEY_minus: Gdk.KEY_KP_Subtract,
            Gdk.KEY_KP_Subtract: Gdk.KEY_minus,
            Gdk.KEY_0: Gdk.KEY_KP_0,
            Gdk.KEY_KP_0: Gdk.KEY_0
        }

        if entry[0] in mapping:
            key = mapping[entry[0]]
            mod = entry[1]

            callback = self._proxy_callback_map[action.get_name()]

            self._accel_group.connect_group(key, mod, Gtk.ACCEL_LOCKED, callback)
            self._proxy_mapping[action] = (key, mod)

    def _init_proxy_accels(self):
        self._install_proxy('LargerTextAction')
        self._install_proxy('SmallerTextAction')
        self._install_proxy('NormalSizeAction')

    def do_deactivate(self):
        # Remove any installed menu items
        self._remove_menu()

        for view in self.window.get_views():
            self.remove_document_helper(view)

        self.window.remove_accel_group(self._accel_group)

        Gtk.AccelMap.get().disconnect(self._accel_map_handler_id)

        self._accel_group = None
        self._action_group = None

    def _insert_menu(self):
        # Get the GtkUIManager
        manager = self.window.get_ui_manager()

        # Create a new action group
        self._action_group = Gtk.ActionGroup("GeditTextSizePluginActions")
        self._action_group.add_actions([("LargerTextAction", None, _("_Larger Text"),
                                         "<Ctrl>plus", None,
                                         self.on_larger_text_activate),
                                         ("SmallerTextAction", None, _("S_maller Text"),
                                         "<Ctrl>minus", None,
                                         self.on_smaller_text_activate),
                                         ("NormalSizeAction", None, _("_Normal size"),
                                         "<Ctrl>0", None,
                                         self.on_normal_size_activate)])

        # Insert the action group
        manager.insert_action_group(self._action_group)

        # Merge the UI
        self._ui_id = manager.add_ui_from_string(ui_str)

    def _remove_menu(self):
        # Get the GtkUIManager
        manager = self.window.get_ui_manager()

        # Remove the ui
        manager.remove_ui(self._ui_id)

        # Remove the action group
        manager.remove_action_group(self._action_group)

        # Make sure the manager updates
        manager.ensure_update()

    def do_update_state(self):
        self._action_group.set_sensitive(self.window.get_active_document() != None)

    def get_helper(self, view):
        if not hasattr(view, "textsize_document_helper"):
            return None
        return view.textsize_document_helper

    def add_document_helper(self, view):
        if self.get_helper(view) != None:
            return

        DocumentHelper(view)

    def remove_document_helper(self, view):
        helper = self.get_helper(view)

        if helper != None:
            helper.stop()

    def call_helper(self, cb):
        view = self.window.get_active_view()

        if view:
            cb(self.get_helper(view))

    # Menu activate handlers
    def on_larger_text_activate(self, action, user_data=None):
        self.call_helper(lambda helper: helper.larger_text())

    def on_smaller_text_activate(self, action, user_data=None):
        self.call_helper(lambda helper: helper.smaller_text())

    def on_normal_size_activate(self, action, user_data=None):
        self.call_helper(lambda helper: helper.normal_size())

    def on_larger_text_accel(self, group, accel, key, mod):
        self.call_helper(lambda helper: helper.larger_text())

    def on_smaller_text_accel(self, group, accel, key, mod):
        self.call_helper(lambda helper: helper.smaller_text())

    def on_normal_size_accel(self, group, accel, key, mod):
        self.call_helper(lambda helper: helper.normal_size())

    def on_tab_added(self, window, tab):
        self.add_document_helper(tab.get_view())

    def on_tab_removed(self, window, tab):
        self.remove_document_helper(tab.get_view())

    def _remap_proxy(self, action):
        # Remove previous proxy

        if action in self._proxy_mapping:
            item = self._proxy_mapping[action]
            self._accel_group.disconnect_key(item[0], item[1])

        self._install_proxy(action)

    def on_accel_map_changed(self, accelmap, path, key, mod):
        for action in self._action_group.list_actions():
            if action.get_accel_path() == path:
                self._remap_proxy(action)
                return

# ex:ts=4:et:



# -*- coding: utf-8 -*-

#  Copyright (C) 2013 - Ignacio Casal Quinteiro
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330,
#  Boston, MA 02111-1307, USA.

from gi.repository import Gdk, Gtk, GtkSource


class DiffType:
    (NONE,
     ADDED,
     MODIFIED,
     REMOVED) = range(4)


class DiffRenderer(GtkSource.GutterRenderer):

    backgrounds = {}
    backgrounds[DiffType.ADDED] = Gdk.RGBA()
    backgrounds[DiffType.MODIFIED] = Gdk.RGBA()
    backgrounds[DiffType.REMOVED] = Gdk.RGBA()
    backgrounds[DiffType.ADDED].parse("#8ae234")
    backgrounds[DiffType.MODIFIED].parse("#fcaf3e")
    backgrounds[DiffType.REMOVED].parse("#ef2929")

    def __init__(self):
        GtkSource.GutterRenderer.__init__(self)

        self.set_size(8)
        self.set_padding(3, 0)

        self.file_context = {}
        self.tooltip = None
        self.tooltip_line = 0

    def do_draw(self, cr, bg_area, cell_area, start, end, state):
        GtkSource.GutterRenderer.do_draw(self, cr, bg_area, cell_area,
                                         start, end, state)

        line_context = self.file_context.get(start.get_line() + 1, None)
        if line_context is None or line_context.line_type == DiffType.NONE:
            return

        background = self.backgrounds[line_context.line_type]

        Gdk.cairo_set_source_rgba(cr, background)
        cr.rectangle(cell_area.x, cell_area.y,
                     cell_area.width, cell_area.height)
        cr.fill()

    def do_query_tooltip(self, it, area, x, y, tooltip):
        line = it.get_line() + 1

        line_context = self.file_context.get(line, None)
        if line_context is None:
            return False

        # Check that the context is the same not the line this
        # way contexts that span multiple times are handled correctly
        if self.file_context.get(self.tooltip_line, None) is line_context:
            tooltip.set_custom(None)
            tooltip.set_custom(self.tooltip)
            return True

        if line_context.line_type not in (DiffType.REMOVED, DiffType.MODIFIED):
            return False

        tooltip_buffer = GtkSource.Buffer()
        tooltip_view = GtkSource.View.new_with_buffer(tooltip_buffer)

        # Propagate the view's settings
        content_view = self.get_view()
        tooltip_view.set_indent_width(content_view.get_indent_width())
        tooltip_view.set_tab_width(content_view.get_tab_width())

        # Propagate the buffer's settings
        content_buffer = content_view.get_buffer()
        tooltip_buffer.set_highlight_syntax(content_buffer.get_highlight_syntax())
        tooltip_buffer.set_language(content_buffer.get_language())
        tooltip_buffer.set_style_scheme(content_buffer.get_style_scheme())

        # Fix some styling issues
        tooltip_buffer.set_highlight_matching_brackets(False)
        tooltip_view.set_border_width(4)
        tooltip_view.set_cursor_visible(False)

        # Set the font
        content_style_context = content_view.get_style_context()
        content_font = content_style_context.get_font(Gtk.StateFlags.NORMAL)
        tooltip_view.override_font(content_font)

        # Only add what can be shown, we
        # don't want to add hundreds of lines
        allocation = content_view.get_allocation()
        lines = allocation.height // area.height
        removed = '\n'.join(map(str, line_context.removed_lines[:lines]))
        tooltip_buffer.set_text(removed)

        # Avoid having to create the tooltip multiple times
        self.tooltip = tooltip_view
        self.tooltip_line = line

        tooltip.set_custom(tooltip_view)
        return True

    def set_file_context(self, file_context):
        self.file_context = file_context
        self.tooltip = None
        self.tooltip_line = 0

        self.queue_draw()

# ex:ts=4:et:



# -*- coding: utf-8 -*-
#
#  rollbackimporter.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import sys
import utils

class RollbackImporter:
    def __init__(self):
        "Creates an instance and installs as the global importer"
        self._new_modules = []
        self._original_import = __builtins__['__import__']

    def monitor(self):
        __builtins__['__import__'] = self._import

    def cancel(self):
        __builtins__['__import__'] = self._original_import

    def _import(self, name, globals={}, locals={}, fromlist=[], level=0):
        maybe = not name in sys.modules

        mod = self._original_import(name, globals, locals, fromlist, level)

        if maybe and utils.is_commander_module(mod):
            self._new_modules.append(name)

        return mod

    def uninstall(self):
        self.cancel()

        for modname in self._new_modules:
            if modname in sys.modules:
                del sys.modules[modname]

        self._new_modules = []

# vi:ex:ts=4:et



#!/usr/bin/env python
# -- coding: utf-8 --
#
# Copyright © 2011 Collabora Ltd.
#            By Seif Lotfy <seif@lotfy.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, Gedit, Gtk, Gio, GLib, GdkPixbuf, GtkSource
from gi.repository import Zeitgeist
from .utils import *

import time
import os
import urllib
import dbus
import datetime

CLIENT = Zeitgeist.Log.get_default()

try:
    BUS = dbus.SessionBus()
    obj = BUS.get_object('org.gnome.zeitgeist.SimpleIndexer',
        '/org/gnome/zeitgeist/index/activity')
    iface = dbus.Interface(obj, 'org.gnome.zeitgeist.Index')
    ZG_FTS = Zeitgeist.Index.new()
except Exception as err:
    ZG_FTS = None
    print ("Could not detect FTS supports, disabling Dashboard search")

version = [int(x) for x in CLIENT.get_version()]

MIN_VERSION = [0, 9, 11, 0]
if version < MIN_VERSION:
    print("PLEASE USE ZEITGEIST 0.9.11 or above")

class Item(Gtk.Button):

    def __init__(self, subject):

        Gtk.Button.__init__(self)
        self._file_object = Gio.file_new_for_uri(subject.get_property("uri"))
        self._file_info =\
            self._file_object.query_info("standard::content-type",
            Gio.FileQueryInfoFlags.NONE, None)
        self.subject = subject
        SIZE_LARGE = (256, 256)
        self.thumb = create_text_thumb(self, SIZE_LARGE, 1)
        self.set_size_request(256, 256)
        vbox = Gtk.Box(orientation = Gtk.Orientation.VERTICAL)
        vbox.pack_start(self.thumb, True, True, 0)
        self.label = Gtk.Label()
        self.label.set_ellipsize(2)
        self.label.set_markup("<span size='large'><b>%s</b></span>"\
            %subject.get_property("text"))
        vbox.pack_start(self.label, False, False, 6)
        self.add(vbox)

    @property
    def mime_type(self):
        return self._file_info.get_attribute_string("standard::content-type")

    def get_content(self):
        f = open(self._file_object.get_path())
        try:
            content = f.read()
        finally:
            f.close()
        return content


class StockButton(Gtk.Button):

    def __init__(self, stock, label):
        Gtk.Button.__init__(self)
        self.set_size_request(256, 256)
        vbox = Gtk.Box(orientation = Gtk.Orientation.VERTICAL)
        self.label = Gtk.Label()
        self.label.set_markup("<span size='large'><b>%s</b></span>"%label)
        vbox.pack_end(self.label, False, False, 6)
        self.add(vbox)


class DashView(Gtk.Box):

    def __init__(self, show_doc, window):
        Gtk.Box.__init__(self, orientation = Gtk.Orientation.VERTICAL)
        hbox = Gtk.Box()
        self._window = window
        self.pack_start(Gtk.Label(""), True, True, 0)
        self.pack_start(hbox, False, False, 0)
        self.pack_start(Gtk.Label(""), True, True, 0)

        self.grid = Gtk.Grid()
        self.grid.set_row_homogeneous(True)
        self.grid.set_column_homogeneous(True)

        self.grid.set_column_spacing(15)
        self.grid.set_row_spacing(15)
        hbox.pack_start(Gtk.Label(""), True, True, 0)
        hbox.pack_start(self.grid, False, False, 0)
        hbox.pack_start(Gtk.Label(""), True, True, 0)

        self.new_button = StockButton(Gtk.STOCK_NEW, _("Empty Document"))
        self.new_button.connect("clicked", lambda x: show_doc())

    def populate_grid(self, subjects):
        for child in self.grid.get_children():
            self.grid.remove(child)
        self.hide()

        self.grid.add(self.new_button)
        self.show_all()

        for i, subject in enumerate(subjects):
            item = Item(subjects[i])
            item.connect("clicked", self.open)
            self.grid.attach(item, (i + 1) % 4, (i + 1) / 4, 1, 1)
        self.show_all()

    def open(self, item):
        Gedit.commands_load_location(self._window,
            item._file_object, None, -1, -1)


class SearchEntry(Gtk.Entry):

    __gsignals__ = {
        "clear": (GObject.SIGNAL_RUN_FIRST,
                   GObject.TYPE_NONE,
                   ()),
        "search": (GObject.SIGNAL_RUN_FIRST,
                    GObject.TYPE_NONE,
                    (GObject.TYPE_STRING,)),
        "close": (GObject.SIGNAL_RUN_FIRST,
                   GObject.TYPE_NONE,
                   ()),
    }

    search_timeout = 0

    def __init__(self, accel_group = None):
        Gtk.Entry.__init__(self)
        self.set_width_chars(40)
        self.set_placeholder_text(_("Type here to search..."))
        self.connect("changed", lambda w: self._queue_search())

        search_icon =\
            Gio.ThemedIcon.new_with_default_fallbacks("edit-find-symbolic")
        self.set_icon_from_gicon(Gtk.EntryIconPosition.PRIMARY, search_icon)
        clear_icon =\
            Gio.ThemedIcon.new_with_default_fallbacks("edit-clear-symbolic")
        self.set_icon_from_gicon(Gtk.EntryIconPosition.SECONDARY, clear_icon)
        self.connect("icon-press", self._icon_press)
        self.show_all()

    def _icon_press(self, widget, pos, event):
        if event.button == 1 and pos == 1:
            self.set_text("")
            self.emit("clear")

    def _queue_search(self):
        if self.search_timeout != 0:
            GObject.source_remove(self.search_timeout)
            self.search_timeout = 0

        if len(self.get_text()) == 0:
            self.emit("clear")
        else:
            self.search_timeout = GObject.timeout_add(200,
                self._typing_timeout)

    def _typing_timeout(self):
        if len(self.get_text()) > 0:
            self.emit("search", self.get_text())

        self.search_timeout = 0
        return False


class ListView(Gtk.TreeView):

    def __init__(self):
        Gtk.TreeView.__init__(self)
        self.icontheme = Gtk.IconTheme.get_default()
        self.model = Gtk.ListStore(str, GdkPixbuf.Pixbuf, str, str,
            GObject.TYPE_PYOBJECT)

        renderer_text = Gtk.CellRendererText()
        column_text = Gtk.TreeViewColumn("Text", renderer_text, markup=0)
        self.append_column(column_text)

        renderer_pixbuf = Gtk.CellRendererPixbuf()
        column_pixbuf = Gtk.TreeViewColumn("Image", renderer_pixbuf, pixbuf=1)
        self.append_column(column_pixbuf)

        renderer_text = Gtk.CellRendererText()
        renderer_text.set_alignment(0.0, 0.5)
        renderer_text.set_property('ellipsize', 1)
        column_text = Gtk.TreeViewColumn("Text", renderer_text, markup=2)
        column_text.set_expand(True)
        self.append_column(column_text)

        renderer_text = Gtk.CellRendererText()
        renderer_text.set_alignment(1.0, 0.5)
        column_text = Gtk.TreeViewColumn("Text", renderer_text, markup=3)
        self.append_column(column_text)

        self.style_get_property("horizontal-separator", 9)
        self.set_headers_visible(False)
        self.set_model(self.model)
        self.home_path = "file://%s" %os.getenv("HOME")
        self.now = time.time()

    def get_time_string(self, timestamp):
        timestamp = int(timestamp)/1000
        diff = int(self.now - timestamp)
        if diff < 60:
            return "   few moment ago   "
        diff = int(diff / 60)
        if diff < 60:
            if diff == 1:
                return "   %i minute ago   " %int(diff)
            return "   %i minutes ago   " %int(diff)
        diff = int(diff / 60)
        if diff < 24:
            if diff == 1:
                return "   %i hour ago   " %int(diff)
            return "   %i hours ago   " %int(diff)
        diff = int(diff / 24)
        if diff < 4:
            if diff == 1:
                return "   Yesterday   "
            return "   %i days ago   " %int(diff)

        return datetime.datetime.fromtimestamp(timestamp)\
            .strftime('   %d %B %Y   ')

    def clear(self):
        self.model.clear()
        self.now = time.time()

    def insert_results(self, log, res, data):
        events = log.search_finish(res)
        self.clear()
        for i in range(events.size()):
            event = events.next_value()
            self.add_item(event, 48)

    def add_item(self, event, icon_size):
        item = event.get_subject(0)
        if uri_exists(item.get_property("uri")):
            uri = item.get_property("uri").replace(self.home_path, "Home").split("/")
            uri = " → ".join(uri[:-1])
            text = """<span size='large'><b>
               %s\n</b></span><span color='darkgrey'>   %s
               </span>"""%(item.get_property("text"), uri)
            iter = self.model.append(["   ", no_pixbuf, text,
                "<span color='darkgrey'>   %s</span>"\
                %self.get_time_string(event.get_property("timestamp")), item])

            def callback(icon):
                self.model[iter][1] = icon if icon is not None else no_pixbuf
            get_icon(item, callback, icon_size)


class Dashboard (Gtk.Box):

    def __init__(self, window, show_doc):
        Gtk.Box.__init__(self, orientation = Gtk.Orientation.VERTICAL)
        self._window = window
        self._show_doc = show_doc
        self._init_done = False
        self.search = SearchEntry()

        if ZG_FTS:
            box = Gtk.Box()
            box.set_homogeneous(True)
            box.pack_start(Gtk.Label(), True, True, 0)
            box.pack_start(self.search, True, True, 0)
            box.pack_start(Gtk.Label(), True, True, 0)
            self.search.connect("search", self._on_search)
            self.search.connect("clear", self._on_clear)
            self.connect("draw", self.change_style)
            self.pack_start(box, False, False, 16)

        self.view = DashView(self._show_doc, window)
        scrolledwindow = Gtk.ScrolledWindow()
        vbox = Gtk.Box(orientation = Gtk.Orientation.VERTICAL)
        scrolledwindow.add(vbox)
        vbox.pack_start(self.view, True, True, 0)
        self.tree_view = ListView()
        vbox.pack_start(self.tree_view, True, True, 0)

        self.pack_start(scrolledwindow, True, True, 0)
        self.get_recent()
        self.show_all()

        def open_uri(widget, row, cell):
            Gedit.commands_load_location(self._window,
                Gio.file_new_for_uri(widget.model[row][4].get_property("uri")),
                    None, 0, 0)
        self.tree_view.connect("row-activated", open_uri)
        self.grab_focus()


    def grab_focus(self):
        if ZG_FTS:
            self.search.grab_focus()

    def _on_search(self, widget, query):
        self.tree_view.show()
        self.view.hide()
        print("Dashboard search for:", query)
        result_type_relevancy = 100
        self.template = Zeitgeist.Event()
        self.template.set_property("actor", "application://gedit.desktop")
        timerange = Zeitgeist.TimeRange.anytime()
        ZG_FTS.search(query + "*",
            timerange,
            [self.template],
            0, 100, 2, None, self.tree_view.insert_results, None)

    def _on_clear(self, widget):
        self.tree_view.hide()
        self.view.show()
        self.tree_view.clear()

    def change_style(self, widget, data):
        self.style = self.get_style_context()
        if self._init_done == False and ZG_FTS:
            self.search.grab_focus()
            self._init_done = True

    def get_recent(self):
        template = Zeitgeist.Event()
        template.set_property("actor", "application://gedit.desktop")
        CLIENT.find_events(
            Zeitgeist.TimeRange.anytime(),
            [template],
            Zeitgeist.StorageState.ANY,
            100,
            Zeitgeist.ResultType.MOST_RECENT_SUBJECTS,
            None,
            self.get_frequent,
            None)

    def get_frequent(self, log, res, data):
        events = CLIENT.find_events_finish(res)
        template = Zeitgeist.Event()
        template.set_property("actor", "application://gedit.desktop")
        now = time.time() * 1000
        # 14 being the amount of days
        # and 86400000 the amount of milliseconds per day
        two_weeks_in_ms = 14 * 86400000
        timerange = Zeitgeist.TimeRange.new(now - two_weeks_in_ms, now)
        CLIENT.find_events(
            timerange,
            [template],
            Zeitgeist.StorageState.ANY,
            100,
            Zeitgeist.ResultType.MOST_POPULAR_SUBJECTS,
            None,
            self.validate_results,
            events)

    def validate_results(self, log, res, data):
        recent_events = data
        popular_events = CLIENT.find_events_finish(res)
        subjects = []
        for events in (recent_events, popular_events,):
            allowed_len = 4 if events == recent_events else 7
            for i in range(events.size()):
                if len(subjects) == allowed_len:
                    break
                event = events.next_value()
                for i in range(event.num_subjects()):
                    subj = event.get_subject(i)
                    if uri_exists(subj.get_property("uri")):
                        subjects.append(subj)
                    if len(subjects) == allowed_len:
                        break
                if len(subjects) == allowed_len:
                    break
        self.view.populate_grid(subjects)

# ex:ts=4:et:



# -*- coding: utf-8 -*-   

#  synctex.py - Synctex support with Gedit and Evince.
#  
#  Copyright (C) 2010 - José Aliste <jose.aliste@gmail.com>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#   
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#   
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, Pango, Gtk, Gedit, Peas, PeasGtk, Gio, Gdk
from .evince_dbus import EvinceWindowProxy
import dbus.mainloop.glib
import logging
import gettext
import os
from gpdefs import *

try:
    gettext.bindtextdomain(GETTEXT_PACKAGE, GP_LOCALEDIR)
    _ = lambda s: gettext.dgettext(GETTEXT_PACKAGE, s);
except:
    _ = lambda s: s

dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

ui_str = """<ui>
  <menubar name="MenuBar">
    <menu name="ToolsMenu" action="Tools">
      <placeholder name="ToolsOps_2">
        <separator/>
        <menuitem name="Synctex" action="SynctexForwardSearch"/>
      </placeholder>
    </menu>
  </menubar>
</ui>
"""

_logger = logging.getLogger("SynctexPlugin")

def apply_style (style, tag):
    def apply_style_prop(tag, style, prop):
        if style.get_property(prop + "-set"):
            tag.set_property(prop, style.get_property(prop))
        else:
            tag.set_property(prop, None)

    def apply_style_prop_bool(tag, style, prop, whentrue, whenfalse):
        if style.get_property(prop + "-set"):
            prop_value = whentrue if style.get_property(prop) else whenfalse
            tag.set_property(prop, prop_value)

    apply_style_prop(tag, style, "foreground")
    apply_style_prop(tag, style, "background")

    apply_style_prop_bool(tag, style, "bold", Pango.Weight.BOLD, Pango.Weight.NORMAL)
    apply_style_prop_bool(tag, style, "italic", Pango.Style.ITALIC, Pango.Style.NORMAL)
    apply_style_prop_bool(tag, style, "underline", Pango.Underline.SINGLE,
                          Pango.Underline.NONE)
    apply_style_prop(tag, style, "strikethrough")

def parse_modeline(line):
    index = line.find("mainfile:")

    if line.startswith("%") and index > 0:
        # Parse the modeline.
        return line[index + len("mainfile:"):].strip()

    return None

class SynctexViewHelper:
    def __init__(self, view, window, plugin):
        self._view = view
        self._window = window
        self._plugin = plugin
        self._doc = view.get_buffer()
        self.window_proxy = None

        self._handlers = [
            self._doc.connect('saved', self.on_saved_or_loaded),
            self._doc.connect('loaded', self.on_saved_or_loaded)
        ]

        self._highlight_tag = self._doc.create_tag()
        self.active = False
        self.last_iters = None
        self.gfile = None
        self.update_location()

    def on_notify_style_scheme(self, doc, param_object):
        apply_style (doc.get_style_scheme().get_style('search-match'), self._highlight_tag)

    def on_button_release(self, view, event):
        modifier_mask = Gtk.accelerator_get_default_mod_mask()
        event_state = event.state & modifier_mask

        if event.button == 1 and event_state == Gdk.ModifierType.CONTROL_MASK:
            self.sync_view(event.time)

    def on_saved_or_loaded(self, doc, data):
        self.update_location()

    def get_output_file(self):
        file_output = None
        line_count = self._doc.get_line_count()

        for i in list(range(min(3,line_count))) + list(range(max(0,line_count - 3), line_count)):
            start = self._doc.get_iter_at_line(i)
            end = start.copy()
            end.forward_to_line_end()
            file_output = parse_modeline(self._doc.get_text(start, end, False))
            if file_output is not None:
                break

        return file_output

    def on_key_press(self, a, b):
        self._unhighlight()

    def on_cursor_moved(self, cur):
        self._unhighlight()

    def deactivate(self):
        self._unhighlight()

        for h in self._handlers:
            self._doc.disconnect(h)

        del self._highlight_tag

    def update_location(self):
        gfile = self._doc.get_location()

        if gfile is None:
            return

        if self.gfile is None or gfile.get_uri() != self.gfile.get_uri():
            SynctexWindowActivatable.view_dict[gfile.get_uri()] = self
            self.gfile = gfile

        modeline_output_file = self.get_output_file()

        if modeline_output_file is not None:
            filename = modeline_output_file
        else:
            filename = self.gfile.get_basename()

        out_path = self.gfile.get_parent().get_child(filename).get_path()
        out_path = os.path.splitext(out_path)
        out_gfile = Gio.file_new_for_path(out_path[0] + ".pdf")

        if out_gfile.query_exists(None):
            self.out_gfile = out_gfile
        else:
            self.out_gfile = None

        self.update_active()

    def _highlight(self):
        iter = self._doc.get_iter_at_mark(self._doc.get_insert())
        end_iter = iter.copy()
        end_iter.forward_to_line_end()

        self._doc.apply_tag(self._highlight_tag, iter, end_iter)
        self.last_iters = [iter, end_iter];

    def _unhighlight(self):
        if self.last_iters is not None:
            self._doc.remove_tag(self._highlight_tag,
                                 self.last_iters[0], self.last_iters[1])
        self.last_iters = None

    def goto_line (self, line, time):
        self._doc.goto_line(line) 
        self._view.scroll_to_cursor()
        self._window.set_active_tab(Gedit.Tab.get_from_document(self._doc))
        self._highlight()
        self._window.present_with_time (time)

    def goto_line_after_load(self, a, line, time):
        GObject.idle_add (lambda : self.goto_line(line, time))
        self._doc.disconnect(self._goto_handler)

    def sync_view(self, time):
        if self.active:
            cursor_iter =  self._doc.get_iter_at_mark(self._doc.get_insert())
            line = cursor_iter.get_line() + 1
            col = cursor_iter.get_line_offset()
            self.window_proxy.SyncView(self.gfile.get_path(), (line, col), time)

    def update_active(self):
        # Activate the plugin only if the doc is a LaTeX file.
        lang = self._doc.get_language()
        self.active = (lang is not None and lang.get_id() == 'latex' and
                        self.out_gfile is not None)

        if self.active and self.window_proxy is None:
            self._doc_active_handlers = [
                        self._doc.connect('cursor-moved', self.on_cursor_moved),
                        self._doc.connect('notify::style-scheme', self.on_notify_style_scheme)]
            self._view_active_handlers = [
                        self._view.connect('key-press-event', self.on_key_press),
                        self._view.connect('button-release-event', self.on_button_release)]


            style = self._doc.get_style_scheme().get_style('search-match')
            apply_style(style, self._highlight_tag)
            self._plugin._action_group.set_sensitive(True)
            self.window_proxy = self._plugin.ref_evince_proxy(self.out_gfile, self._window)

        elif not self.active and self.window_proxy is not None:
            #destroy the evince window proxy.
            for handler in self._doc_active_handlers:
                self._doc.disconnect(handler)
            for handler in self._view_active_handlers:
                self._view.disconnect(handler)

            self._plugin._action_group.set_sensitive(False)
            self._plugin.unref_evince_proxy(self.out_gfile)
            self.window_proxy = None


class SynctexWindowActivatable(GObject.Object, Gedit.WindowActivatable):
    __gtype_name__ = "SynctexWindowActivatable"

    window = GObject.property(type=Gedit.Window)
    view_dict = {}
    _proxy_dict = {}

    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        self._insert_menu()

        for view in self.window.get_views():
            self.add_helper(view, self.window)

        self.handlers = [
            self.window.connect("tab-added", lambda window, tab: self.add_helper(tab.get_view(), window)),
            self.window.connect("tab-removed", lambda window, tab: self.remove_helper(tab.get_view())),
            self.window.connect("active-tab-changed", self.on_active_tab_changed)
        ]

    def do_deactivate(self):
        for h in self.handlers:
            self.window.disconnect(h)

        for view in self.window.get_views():
            self.remove_helper(view)

        self._remove_menu()

    def on_active_tab_changed(self, window,  tab):
        view_helper = self.get_helper(tab.get_view())

        if view_helper is None:
            active = False
        else:
            active = view_helper.active

        self._action_group.set_sensitive(active)

    def add_helper(self, view, window):
        helper = SynctexViewHelper(view, window, self)
        location = view.get_buffer().get_location()

        if location is not None:
            self.view_dict[location.get_uri()] = helper
        view.synctex_view_helper = helper

    def remove_helper(self, view):
        helper = self.get_helper(view)

        if helper.gfile is not None:
            del self.view_dict[helper.gfile.get_uri()]

        helper.deactivate()
        del view.synctex_view_helper

    def get_helper(self, view):
        if not hasattr(view, 'synctex_view_helper'):
            return None
        return view.synctex_view_helper

    def _remove_menu(self):
        manager = self.window.get_ui_manager()
        manager.remove_ui(self._ui_id)
        manager.remove_action_group(self._action_group)
        manager.ensure_update()

    def _insert_menu(self):
        # Get the GtkUIManager
        manager = self.window.get_ui_manager()

        # Create a new action group
        self._action_group = Gtk.ActionGroup(name="SynctexWindowActivatableActions")
        self._action_group.add_actions([("SynctexForwardSearch", None,
                                        _("Forward Search"), "<Ctrl><Alt>F",
                                        _("Forward Search"), self.forward_search_cb)])

        # Insert the action group
        manager.insert_action_group(self._action_group, -1)

        # Merge the UI
        self._ui_id = manager.add_ui_from_string(ui_str)

    def forward_search_cb(self, action, what=None):
        self.get_helper(self.window.get_active_view()).sync_view(Gtk.get_current_event_time())

    def source_view_handler(self, out_gfile, uri_input, source_link, time):

        if uri_input not in self.view_dict:
            window = self._proxy_dict[out_gfile.get_uri()][2]

            tab = window.create_tab_from_location(Gio.file_new_for_uri(uri_input),
                                                  None, source_link[0] - 1, 0, False, True)

            helper = self.get_helper(tab.get_view())
            helper._goto_handler = tab.get_document().connect_object("loaded", 
                                                SynctexViewHelper.goto_line_after_load,
                                                helper, source_link[0] - 1, time)
        else:
            self.view_dict[uri_input].goto_line(source_link[0] - 1, time)

    def ref_evince_proxy(self, gfile, window):
        uri = gfile.get_uri()
        proxy = None

        if uri not in self._proxy_dict:
            proxy = EvinceWindowProxy (uri, True, _logger)
            self._proxy_dict[uri] = [1, proxy, window]
            proxy.set_source_handler (lambda i, s, time: self.source_view_handler(gfile, i, s, time))
        else:
            self._proxy_dict[uri][0]+=1
            proxy = self._proxy_dict[uri][1]

        return proxy

    def unref_evince_proxy(self, gfile):
        uri = gfile.get_uri()

        if uri in self._proxy_dict:
            self._proxy_dict[uri][0] -= 1
            if self._proxy_dict[uri][0] == 0:
                del self._proxy_dict[uri]

# ex:ts=4:et:



#!/usr/bin/python
#
# Gedit Scheme Editor
# https://github.com/jonocodes/GeditSchemer
#
# Copyright (C) Jono Finger 2013 <jono@foodnotblogs.com>
# 
# The program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
# 
# The program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, Gedit, Gtk
import os

from .schemer import GUI

UI_XML = """<ui>
<menubar name="MenuBar">
  <menu name="ToolsMenu" action="Tools">
    <placeholder name="ToolsOps_4">
      <menuitem name="menuItemLaunchGui" action="LaunchGuiAction"/>
    </placeholder>
  </menu>
</menubar>
</ui>"""

class WindowActivatable(GObject.Object, Gedit.WindowActivatable):

  window = GObject.property(type=Gedit.Window)

  def __init__(self):
    GObject.Object.__init__(self)

  def do_activate(self):
    manager = self.window.get_ui_manager()
    self._actions = Gtk.ActionGroup("SchemerActions")
    self._actions.add_actions([
      ('LaunchGuiAction', Gtk.STOCK_INFO, "Color Scheme Editor", 
        None, "Launch color scheme editor for the current loaded scheme", 
        self.open_dialog),
    ])
    manager.insert_action_group(self._actions)
    self._ui_merge_id = manager.add_ui_from_string(UI_XML)
    manager.ensure_update()

  def open_dialog(self, action, data=None):
    schemer.GUI(Gedit.App, os.path.join(self.plugin_info.get_data_dir(), 'ui'))

  def do_deactivate(self):
    manager = self.window.get_ui_manager()
    manager.remove_ui(self._ui_merge_id)
    manager.remove_action_group(self._actions)
    manager.ensure_update()
    



# -*- coding: utf-8 -*-
#
#  multiedit.py - Multi Edit
#
#  Copyright (C) 2009 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, Gtk, Gedit
from .signals import Signals
from .documenthelper import DocumentHelper
import gettext
from gpdefs import *

try:
    gettext.bindtextdomain(GETTEXT_PACKAGE, GP_LOCALEDIR)
    _ = lambda s: gettext.dgettext(GETTEXT_PACKAGE, s);
except:
    _ = lambda s: s

ui_str = """
<ui>
  <menubar name="MenuBar">
    <menu name="EditMenu" action="Edit">
      <placeholder name="EditOps_5">
        <menuitem name="MultiEditMode" action="MultiEditModeAction"/>
      </placeholder>
    </menu>
  </menubar>
</ui>
"""

class MultiEditPlugin(GObject.Object, Gedit.WindowActivatable, Signals):
    __gtype_name__ = "MultiEditPlugin"

    window = GObject.property(type=Gedit.Window)

    def __init__(self):
        GObject.Object.__init__(self)
        Signals.__init__(self)

    def do_activate(self):
        for view in self.window.get_views():
            self.add_document_helper(view)

        self.connect_signal(self.window, 'tab-added', self.on_tab_added)
        self.connect_signal(self.window, 'tab-removed', self.on_tab_removed)
        self.connect_signal(self.window, 'active-tab-changed', self.on_active_tab_changed)

        self._insert_menu()

    def do_deactivate(self):
        self.disconnect_signals(self.window)

        for view in self.window.get_views():
            self.remove_document_helper(view)

        self._remove_menu()

    def _insert_menu(self):
        manager = self.window.get_ui_manager()

        self._action_group = Gtk.ActionGroup("GeditMultiEditPluginActions")
        self._action_group.add_toggle_actions([('MultiEditModeAction',
                                                None,
                                                _('Multi Edit Mode'),
                                                '<Ctrl><Shift>C',
                                                _('Start multi edit mode'),
                                                self.on_multi_edit_mode)])

        manager.insert_action_group(self._action_group)
        self._ui_id = manager.add_ui_from_string(ui_str)

    def _remove_menu(self):
        manager = self.window.get_ui_manager()
        manager.remove_ui(self._ui_id)
        manager.remove_action_group(self._action_group)
        manager.ensure_update()

    def do_update_state(self):
        pass

    def get_helper(self, view):
        if not hasattr(view, "multiedit_document_helper"):
            return None
        return view.multiedit_document_helper

    def add_document_helper(self, view):
        if self.get_helper(view) != None:
            return

        helper = DocumentHelper(view)
        helper.set_toggle_callback(self.on_multi_edit_toggled, helper)

    def remove_document_helper(self, view):
        helper = self.get_helper(view)

        if helper != None:
            helper.stop()

    def get_action(self):
        return self._action_group.get_action('MultiEditModeAction')

    def on_multi_edit_toggled(self, helper):
        if helper.get_view() == self.window.get_active_view():
            self.get_action().set_active(helper.enabled())

    def on_tab_added(self, window, tab):
        self.add_document_helper(tab.get_view())

    def on_tab_removed(self, window, tab):
        self.remove_document_helper(tab.get_view())

    def on_active_tab_changed(self, window, tab):
        view = tab.get_view()
        helper = self.get_helper(view)

        self.get_action().set_active(helper != None and helper.enabled())

    def on_multi_edit_mode(self, action):
        view = self.window.get_active_view()
        helper = self.get_helper(view)

        if helper != None:
            helper.toggle_multi_edit(self.get_action().get_active())

# ex:ts=4:et:



# -*- coding: utf-8 -*-
#
#  info.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from transparentwindow import TransparentWindow
from gi.repository import Pango, Gdk, Gtk
import math

class Info(TransparentWindow):
    __gtype_name__ = "CommanderInfo"

    def __init__(self, entry):
        super(Info, self).__init__(Gtk.WindowType.POPUP)

        self._entry = entry
        self._vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=3)

        ev = Gtk.EventBox()
        ev.show()

        self.set_transient_for(entry.get_toplevel())

        self._vw = Gtk.ScrolledWindow()
        self._vw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
        self._vw.show()

        ev.add(self._vw)

        self._text = Gtk.TextView()

        font = self._entry.get_font()
        fgcolor = self._entry.get_foreground_color()
        bgcolor = self.background_color()

        self._text.override_font(font)
        self._text.override_color(Gtk.StateFlags.NORMAL, fgcolor)
        self._text.override_background_color(Gtk.StateFlags.NORMAL, bgcolor)

        ev.override_background_color(Gtk.StateFlags.NORMAL, bgcolor)

        self._text.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)

        buf = self._text.get_buffer()

        buf.connect_after('insert-text', self.on_text_insert_text)
        buf.connect_after('delete-range', self.on_text_delete_range)

        self._text.set_editable(False)

        self._vw.add(self._text)
        self._vbox.pack_end(ev, False, False, 0)
        self._vbox.show()
        self._button_bar = None

        self.add(self._vbox)
        self._text.show()
        self._status_label = None

        self.props.can_focus = False
        self.set_border_width(8)

        self.attach()
        self.show()

        self.max_lines = 10

        self.connect_after('size-allocate', self.on_size_allocate)
        self.connect_after('draw', self.on_draw)

        self._attr_map = {
            Pango.AttrType.STYLE: 'style',
            Pango.AttrType.WEIGHT: 'weight',
            Pango.AttrType.VARIANT: 'variant',
            Pango.AttrType.STRETCH: 'stretch',
            Pango.AttrType.SIZE: 'size',
            Pango.AttrType.FOREGROUND: 'foreground',
            Pango.AttrType.BACKGROUND: 'background',
            Pango.AttrType.UNDERLINE: 'underline',
            Pango.AttrType.STRIKETHROUGH: 'strikethrough',
            Pango.AttrType.RISE: 'rise',
            Pango.AttrType.SCALE: 'scale'
        }

        css = Gtk.CssProvider()
        css.load_from_data(bytes("""
.trough {
    background: transparent;
}
""", 'utf-8'))

        self._vw.get_vscrollbar().get_style_context().add_provider(css, 600)

    def empty(self):
        buf = self._text.get_buffer()
        return buf.get_start_iter().equal(buf.get_end_iter())

    def status(self, text=None):
        if self._status_label == None and text != None:
            self._status_label = Gtk.Label()

            context = self._text.get_style_context()
            state = context.get_state()
            font_desc = context.get_font(state)

            self._status_label.override_font(font_desc)

            self._status_label.override_color(Gtk.StateFlags.NORMAL, context.get_color(Gtk.StateFlags.NORMAL))
            self._status_label.show()
            self._status_label.set_alignment(0, 0.5)
            self._status_label.set_padding(10, 0)
            self._status_label.set_use_markup(True)

            self.ensure_button_bar()
            self._button_bar.pack_start(self._status_label, True, True, 0)

        if text != None:
            self._status_label.set_markup(text)
        elif self._status_label:
            self._status_label.destroy()

            if not self._button_bar and self.empty():
                self.destroy()

    def attrs_to_tags(self, attrs):
        buf = self._text.get_buffer()
        table = buf.get_tag_table()
        ret = []

        for attr in attrs:
            if not attr.type in self._attr_map:
                continue

            if attr.type == Pango.AttrType.FOREGROUND or attr.type == Pango.AttrType.BACKGROUND:
                value = attr.color
            else:
                value = attr.value

            tagname = str(attr.type) + ':' + str(value)

            tag = table.lookup(tagname)

            if not tag:
                tag = buf.create_tag(tagname)
                tag.set_property(self._attr_map[attr.type], value)

            ret.append(tag)

        return ret

    def add_lines(self, line, use_markup=False):
        buf = self._text.get_buffer()

        if not buf.get_start_iter().equal(buf.get_end_iter()):
            line = "\n" + line

        if not use_markup:
            buf.insert(buf.get_end_iter(), line)
            return

        try:
            ret = Pango.parse_markup(line, -1, '\x00')
        except Exception as e:
            print('Could not parse markup:', e)
            buf.insert(buf.get_end_iter(), line)
            return

        # TODO: fix this when pango supports get_iterator
        text = ret[2]
        buf.insert(buf.get_end_iter(), text)

        return

        piter = ret[1].get_iterator()

        while piter:
            attrs = piter.get_attrs()
            begin, end = piter.range()

            tags = self.attrs_to_tags(attrs)
            buf.insert_with_tags(buf.get_end_iter(), text[begin:end], *tags)

            if not piter.next():
                break

    def toomany_lines(self):
        buf = self._text.get_buffer()
        piter = buf.get_start_iter()
        num = 0

        while self._text.forward_display_line(piter):
            num += 1

            if num > self.max_lines:
                return True

        return False

    def contents_changed(self):
        buf = self._text.get_buffer()

        if self.toomany_lines() and (self._vw.get_policy()[1] != Gtk.PolicyType.ALWAYS):
            self._vw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.ALWAYS)

            layout = self._text.create_pango_layout('Some text to measure')
            extents = layout.get_pixel_extents()

            self._vw.set_min_content_height(extents[1].height * self.max_lines)
        elif not self.toomany_lines() and (self._vw.get_policy()[1] == Gtk.PolicyType.ALWAYS):
            self._vw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
            self._vw.set_min_content_height(0)

        if not self.toomany_lines():
            size = self.get_size()
            self.resize(size[0], 1)

    def ensure_button_bar(self):
        if not self._button_bar:
            self._button_bar = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=3)
            self._button_bar.show()
            self._vbox.pack_start(self._button_bar, False, False, 0)

    def add_action(self, stock, callback, data=None):
        image = Gtk.Image.new_from_stock(stock, Gtk.IconSize.MENU)
        image.show()

        image.commander_action_stock_item = (stock, Gtk.IconSize.MENU)

        self.ensure_button_bar()

        ev = Gtk.EventBox()
        ev.set_visible_window(False)
        ev.add(image)
        ev.show()

        self._button_bar.pack_end(ev, False, False, 0)

        ev.connect('button-press-event', self.on_action_activate, callback, data)
        ev.connect('enter-notify-event', self.on_action_enter_notify)
        ev.connect('leave-notify-event', self.on_action_leave_notify)

        ev.connect_after('destroy', self.on_action_destroy)
        return ev

    def on_action_destroy(self, widget):
        if self._button_bar and len(self._button_bar.get_children()) == 0:
            self._button_bar.destroy()
            self._button_bar = None

    def on_action_enter_notify(self, widget, evnt):
        img = widget.get_child()
        img.set_state(Gtk.StateType.PRELIGHT)
        widget.get_window().set_cursor(Gdk.Cursor.new(Gdk.HAND2))

        pix = img.render_icon(*img.commander_action_stock_item)
        img.set_from_pixbuf(pix)

    def on_action_leave_notify(self, widget, evnt):
        img = widget.get_child()
        img.set_state(Gtk.StateType.NORMAL)
        widget.get_window().set_cursor(None)

        pix = img.render_icon(*img.commander_action_stock_item)
        img.set_from_pixbuf(pix)

    def on_action_activate(self, widget, evnt, callback, data):
        if data:
            callback(data)
        else:
            callback()

    def clear(self):
        self._text.get_buffer().set_text('')

    def attach(self):
        vwwnd = self._entry._view.get_window(Gtk.TextWindowType.TEXT)
        origin = vwwnd.get_origin()
        geom = vwwnd.get_geometry()

        margin = 5

        self.realize()

        alloc = self.get_allocation()

        self.move(origin[1], origin[2] + geom[3] - alloc.height)
        self.resize(geom[2] - margin * 2, alloc.height)

    def on_text_insert_text(self, buf, piter, text, length):
        self.contents_changed()

    def on_text_delete_range(self, buf, start, end):
        self.contents_changed()

    def on_size_allocate(self, widget, allocation):
        vwwnd = self._entry.view().get_window(Gtk.TextWindowType.TEXT)
        origin = vwwnd.get_origin()
        geom = vwwnd.get_geometry()

        alloc = self.get_allocation()

        self.move(origin[1] + (geom[2] - alloc.width) / 2, origin[2] + geom[3] - alloc.height)

    def on_draw(self, widget, ct):
        color = self._entry.get_border_color()

        self.background_shape(ct, self.get_allocation())

        ct.set_source_rgba(color.red, color.green, color.blue, color.alpha)
        ct.stroke()

        return False

    def background_shape(self, ct, alloc):
        w = alloc.width
        h = alloc.height

        ct.set_line_width(1)
        radius = 10

        ct.move_to(0.5, h)

        if self.is_composited():
            ct.arc(radius + 0.5, radius + 0.5, radius, math.pi, math.pi * 1.5)
            ct.arc(w - radius - 0.5, radius + 0.5, radius, math.pi * 1.5, math.pi * 2)
        else:
            ct.line_to(0.5, 0)
            ct.line_to(w - 0.5, 0)

        ct.line_to(w - 0.5, h)

    def background_color(self):
        color = self._entry.get_background_color().copy()
        color.alpha = 0.9

        return color

    def on_text_size_allocate(self, widget, alloc):
        pass

# vi:ex:ts=4:et



# -*- coding: utf-8 -*-

#  smartspaces.py
#
#  Copyright (C) 2006 - Steve Frécinaux
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, Gtk, Gdk, GtkSource, Gedit

class SmartSpacesPlugin(GObject.Object, Gedit.ViewActivatable):
    __gtype_name__ = "SmartSpacesPlugin"

    view = GObject.property(type=Gedit.View)

    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        self._handlers = [
            None,
            self.view.connect('notify::editable', self.on_notify),
            self.view.connect('notify::insert-spaces-instead-of-tabs', self.on_notify)
        ]

    def do_deactivate(self):
        for handler in self._handlers:
            if handler is not None:
                self.view.disconnect(handler)

    def update_active(self):
        # Don't activate the feature if the buffer isn't editable or if
        # we're using tabs
        active = self.view.get_editable() and \
                 self.view.get_insert_spaces_instead_of_tabs()

        if active and self._handlers[0] is None:
            self._handlers[0] = self.view.connect('key-press-event',
                                                   self.on_key_press_event)
        elif not active and self._handlers[0] is not None:
            self.view.disconnect(self._handlers[0])
            self._handlers[0] = None

    def on_notify(self, view, pspec):
        self.update_active()

    def get_real_indent_width(self):
        indent_width = self.view.get_indent_width()

        if indent_width < 0:
             indent_width = self.view.get_tab_width()

        return indent_width

    def on_key_press_event(self, view, event):
        # Only take care of backspace and shift+backspace
        mods = Gtk.accelerator_get_default_mod_mask()

        if event.keyval != Gdk.KEY_BackSpace or \
           event.state & mods != 0 and event.state & mods != Gdk.ModifierType.SHIFT_MASK:
            return False

        doc = view.get_buffer()
        if doc.get_has_selection():
            return False

        cur = doc.get_iter_at_mark(doc.get_insert())
        offset = cur.get_line_offset()

        if offset == 0:
            # We're at the begining of the line, so we can't obviously
            # unindent in this case
            return False

        start = cur.copy()
        prev = cur.copy()
        prev.backward_char()

        # If the previous chars are spaces, try to remove
        # them until the previous tab stop
        max_move = offset % self.get_real_indent_width()

        if max_move == 0:
            max_move = self.get_real_indent_width()

        moved = 0
        while moved < max_move and prev.get_char() == ' ':
            start.backward_char()
            moved += 1
            if not prev.backward_char():
                # we reached the start of the buffer
                break

        if moved == 0:
            # The iterator hasn't moved, it was not a space
            return False

        # Actually delete the spaces
        doc.begin_user_action()
        doc.delete(start, cur)
        doc.end_user_action()

        return True

# ex:ts=4:et:
